import logging
import re


def find_report(log, activity, subject):
    """Check if an activity was logged."""
    for record in log.records:
        record_activity = getattr(record, "activity", None)
        if record_activity == activity and str(subject) in str(record.subject):
            return record

    return False


def make_record(activity, subject, context=None, target=None, nesting=0):
    """Create a custom record."""
    return logging.makeLogRecord(
        dict(
            activity=activity,
            subject=subject,
            context=context,
            target=target,
            nesting=nesting,
        )
    )


def match_record(record, **kwargs):
    for key, value in kwargs.items():
        if getattr(record, key) != value:
            return False

    return record


REPORT_REGEX = re.compile(
    r"^\s*(?P<activity>\w+)(?P<spacing>\s+)(?P<content>.+)$", re.I + re.U
)


def match_report(record, message=None, **kwargs):
    """Check if a log entry was created using `report`, and compare."""
    result = REPORT_REGEX.search(record.message)
    if not result:
        return False

    if message is not None and message not in record.message:
        return False

    match = result.groupdict()

    for key, value in kwargs.items():
        if match[key] != value:
            return False

    return match


def ansi_pattern(text):
    return r"({prefix}\[\d+m)+{text}{prefix}\[0m".format(
        text=re.escape(text), prefix="\033"
    )


def ansi_regex(text):
    return re.compile(ansi_pattern(text), re.I)



import os
import shutil
import sys
from pathlib import Path

import virtualenv

from .system.helpers import normalize_run_args, run_with_debug

_DEFAULT_ENV = dict(os.environ)
_DEFAULT_ENV.pop("PYTHONPATH", None)  # ensure isolation
_PROJ_ROOT = str(Path(__file__).parent.parent)


class VirtualEnv:
    CREATE_OPTS = ["--no-setuptools", "--no-wheel"]

    def __init__(
        self,
        name: str,
        root: "os.PathLike[str]",
        opts=CREATE_OPTS,
        python=sys.executable,
        env=_DEFAULT_ENV,
    ):
        self.name = name
        self.root = Path(root)
        self.opts = opts
        self.path = self.root / name
        self.env = env
        self._python = python
        self._bin = self.path / ("Scripts" if os.name == "nt" else "bin")
        self._coverage_installed = False

    def create(self):
        args = [str(self.path), "--python", self._python, *self.opts]
        virtualenv.cli_run(args, env=self.env)

    def ensure_created(self):
        if not self.path.exists():
            self.create()

    def ensure_coverage(self):
        if self._coverage_installed:
            return

        return self.run(
            "pip install coverage --trusted-host pypi.python.org "
            "--trusted-host files.pythonhosted.org --trusted-host pypi.org"
        )
        self._coverage_installed = True

    def exe(self, cmd: str) -> str:
        return shutil.which(cmd, path=str(self._bin)) or shutil.which(cmd) or cmd

    def run(self, *args: str, with_coverage=False, **kwargs):
        self.ensure_created()
        args, kwargs = normalize_run_args(args, {"env": self.env, **kwargs})
        if with_coverage:
            self.ensure_coverage()
            coverage_dir = kwargs.pop("coverage_dir", ".")
            args = [
                *("coverage", "run", "--parallel-mode"),
                *("--rcfile", _coveragerc(coverage_dir)),
                *("--data-file", _coverage_data_file(coverage_dir)),
                *args,
            ]

        return run_with_debug([self.exe(args[0]), *args[1:]], kwargs)


def _coveragerc(coverage_dir) -> str:
    for root in (coverage_dir, os.getenv("TOXINIDIR", _PROJ_ROOT)):
        file = Path(root, ".coveragerc")  # change if configurations are moved
        if file.exists():
            return str(file)

    raise FileNotFoundError(file)


def _coverage_data_file(coverage_dir) -> str:
    for root in (coverage_dir, os.getenv("TOXINIDIR", _PROJ_ROOT)):
        path = Path(root)
        if list(path.glob(".coverage.*")) or (path / ".coverage").exists():
            return str(path / ".coverage")

    raise FileNotFoundError(".coverage (data_file to store coverage)")



"""Functionality to generate and work with the directory structure of a project.

.. versionchanged:: 4.0
   ``Callable[[dict], str]`` and :obj:`string.Template` objects can also be used as file
   contents. They will be called with snek's ``opts`` (:obj:`string.Template` via
   :obj:`~string.Template.safe_substitute`)
"""
from copy import deepcopy
from pathlib import Path
from string import Template
from typing import Callable, Dict, Optional, Tuple, Union, cast

from . import templates
from .file_system import PathLike, create_directory
from .operations import (
    FileContents,
    FileOp,
    ScaffoldOpts,
    create,
    no_overwrite,
    skip_on_update,
)
from .templates import get_template

NO_OVERWRITE = no_overwrite()
SKIP_ON_UPDATE = skip_on_update()


# Sphinx is bad at documenting aliases for the time being... so we repeat the definition

AbstractContent = Union[FileContents, Callable[[ScaffoldOpts], FileContents], Template]
"""*Recipe* for obtaining file contents
::

    Union[FileContents, Callable[[ScaffoldOpts], FileContents], Template]
"""

ResolvedLeaf = Tuple[AbstractContent, FileOp]
"""Complete *recipe* for manipulating a file in disk (not only its contents but also the
file operation::

    Tuple[AbstractContent, FileOp]
"""

ReifiedLeaf = Tuple[FileContents, FileOp]
"""Similar to :obj:`ResolvedLeaf` but with file contents "reified", i.e. an actual
string instead of a "lazy object" (such as a function or template).
"""

Leaf = Union[AbstractContent, ResolvedLeaf]
"""Just the content of the file OR a tuple of content + file operation
::

    Union[AbstractContent, ResolvedLeaf]
"""

# TODO: Replace `dict` when recursive types are processed by mypy
Node = Union[Leaf, dict]
"""Representation of a *file system node* in the project structure (e.g. files,
directories::

    Union[Leaf, Structure]
"""


Structure = Dict[str, Node]
"""The directory tree represented as a (possibly nested) dictionary::

    Structure = Dict[str, Node]
    Node = Union[Leaf, Structure]

The keys indicate the path where a file will be written, while the
value indicates the content.

A nested dictionary represent a nested directory, while :obj:`str`,
:obj:`string.Template` and :obj:`callable` values represent a file to be created.
:obj:`tuple` values are also allowed, and in that case, the first element of the tuple
represents the file content while the second element is a :mod:`snek.operations
<file operation>` (which can be seen as a recipe on how to create a file with the given
content). :obj:`Callable <callable>` file contents are transformed into strings by
calling them with :obj:`snek's option dict as argument
<snek.api.create_structure>`. Similarly, :obj:`string.Template.safe_substitute`
are called with snek's opts.

The top level keys in the dict are file/dir names relative to the project root, while
keys in nested dicts are relative to the parent's key/location.

For example::

    from snek.operations import no_overwrite
    struct: Structure = {
        'namespace': {
            'module.py': ('print("Hello World!")', no_overwrite())
        }
    }

represents a ``namespace/module.py`` file inside the project folder
with content ``print("Hello World!")``, that will be created only if not
present.

Note:
    :obj:`None` file contents are ignored and not created in disk.
"""

ActionParams = Tuple[Structure, ScaffoldOpts]
"""See :obj:`snek.actions.ActionParams`"""


# -------- snek Actions --------


def define_structure(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Creates the project structure as dictionary of dictionaries

    Args:
        struct : previous directory structure (usually and empty dict)
        opts: options of the project

    Returns:
        Project structure and snek's options

    .. versionchanged:: 4.0
       :obj:`string.Template` and functions added directly to the file structure.
    """
    files: Structure = {
        # Tools
        ".gitignore": (get_template("gitignore"), NO_OVERWRITE),
        ".coveragerc": (get_template("coveragerc"), NO_OVERWRITE),
        ".readthedocs.yml": (get_template("rtd_cfg"), NO_OVERWRITE),
        # Project configuration
        "pyproject.toml": (templates.pyproject_toml, NO_OVERWRITE),
        "setup.py": get_template("setup_py"),
        "setup.cfg": (templates.setup_cfg, NO_OVERWRITE),
        "tox.ini": (get_template("tox_ini"), NO_OVERWRITE),
        # Essential docs
        "README.rst": (get_template("readme"), NO_OVERWRITE),
        "LICENSE.txt": (templates.license, NO_OVERWRITE),
        # Code
        "src": {
            opts["package"]: {
                "__init__.py": templates.init,
                "cli.py": (get_template("cli"), SKIP_ON_UPDATE),
            }
        },
        # Tests
        "tests": {
            "conftest.py": (get_template("conftest_py"), NO_OVERWRITE),
            "test_cli.py": (get_template("test_cli"), SKIP_ON_UPDATE),
        },
        # Remaining of the Documentation
        "docs": {
            "conf.py": get_template("sphinx_conf"),
            "index.rst": (get_template("sphinx_index"), NO_OVERWRITE),
            "readme.rst": get_template("sphinx_readme"),
            "Makefile": get_template("sphinx_makefile"),
            "_static": {".gitignore": get_template("gitignore_empty")},
            "requirements.txt": (get_template("rtd_requirements"), NO_OVERWRITE),
        },
    }

    return merge(struct, files), opts


def create_structure(
    struct: Structure, opts: ScaffoldOpts, prefix: Optional[Path] = None
) -> ActionParams:
    """Manifests/reifies a directory structure in the filesystem

    Args:
        struct: directory structure as dictionary of dictionaries
        opts: options of the project
        prefix: prefix path for the structure

    Returns:
        Directory structure as dictionary of dictionaries (similar to input, but only
        containing the files that actually changed) and input options

    Raises:
        TypeError: raised if content type in struct is unknown

    .. versionchanged:: 4.0
       Also accepts :obj:`string.Template` and :obj:`callable` objects as file contents.
    """
    update = opts.get("update") or opts.get("force")
    pretend = opts.get("pretend")

    if prefix is None:
        prefix = cast(Path, opts.get("project_path", "."))
        create_directory(prefix, update, pretend)
    prefix = Path(prefix)

    changed: Structure = {}

    for name, node in struct.items():
        path = prefix / name
        if isinstance(node, dict):
            create_directory(path, update, pretend)
            changed[name], _ = create_structure(node, opts, prefix=path)
        else:
            content, file_op = reify_leaf(node, opts)
            if file_op(path, content, opts):
                changed[name] = content

    return changed, opts


# -------- Auxiliary Functions --------


def resolve_leaf(contents: Leaf) -> ResolvedLeaf:
    """Normalize project structure leaf to be a ``Tuple[AbstractContent, FileOp]``"""
    if isinstance(contents, tuple):
        return contents
    return (contents, create)


def reify_content(content: AbstractContent, opts: ScaffoldOpts) -> FileContents:
    """Make sure content is string (calling :meth:`~object.__call__` or
    :meth:`~string.Template.safe_substitute` with opts if necessary)
    """
    if callable(content):
        return content(opts)
    if isinstance(content, Template):
        return content.safe_substitute(opts)
    return content


def reify_leaf(contents: Leaf, opts: ScaffoldOpts) -> ReifiedLeaf:
    """Similar to :obj:`resolve_leaf` but applies :obj:`reify_content` to the first
    element of the returned tuple.
    """
    file_contents, action = resolve_leaf(contents)
    return (reify_content(file_contents, opts), action)


# -------- Structure Manipulation --------


def modify(
    struct: Structure,
    path: PathLike,
    modifier: Callable[[AbstractContent, FileOp], ResolvedLeaf],
) -> Structure:
    """Modify the contents of a file in the representation of the project tree.

    If the given path does not exist, the parent directories are automatically
    created.

    Args:
        struct: project representation as (possibly) nested dict. See :obj:`~.merge`.

        path: path-like string or object relative to the structure root.
            The following examples are equivalent::

                from pathlib import Path

                'docs/api/index.html'
                Path('docs', 'api', 'index.html')

            .. versionchanged:: 4.0
               The function no longer accepts a list of strings of path parts.

        modifier: function (or callable object) that receives the
            old content and the old file operation as arguments and returns
            a tuple with the new content and new file operation.
            Note that, if the file does not exist in ``struct``, ``None`` will
            be passed as argument. Example::

                modifier = lambda old, op: ((old or '') + 'APPENDED CONTENT'!, op)
                modifier = lambda old, op: ('PREPENDED CONTENT!' + (old or ''), op)

            .. versionchanged:: 4.0
               ``modifier`` requires 2 arguments and now is a mandatory argument.

    .. versionchanged:: 4.0
       ``update_rule`` is no longer an argument. Instead the arity ``modifier`` was
       changed to accept 2 arguments instead of only 1. This is more suitable to
       handling the new :obj:`snek.operations` API.

    Returns:
        Updated project tree representation

    Note:
        Use an empty string as content to ensure a file is created empty
        (``None`` contents will not be created).
    """
    # Retrieve a list of parts from a path-like object
    path_parts = Path(path).parts

    # Walk the entire path, creating parents if necessary.
    root = deepcopy(struct)
    last_parent: dict = root
    name = path_parts[-1]
    for parent in path_parts[:-1]:
        last_parent = last_parent.setdefault(parent, {})

    # Get the old value if existent.
    old_value = resolve_leaf(last_parent.get(name))

    # Update the value.
    new_value = modifier(*old_value)
    last_parent[name] = _merge_leaf(old_value, new_value)

    return root


def ensure(
    struct: Structure,
    path: PathLike,
    content: AbstractContent = None,
    file_op: FileOp = create,
) -> Structure:
    """Ensure a file exists in the representation of the project tree
    with the provided content.
    All the parent directories are automatically created.

    Args:
        struct: project representation as (possibly) nested.

        path: path-like string or object relative to the structure root.
            See :obj:`~.modify`.

            .. versionchanged:: 4.0
               The function no longer accepts a list of strings of path parts.

        content: file text contents, ``None`` by default.
            The old content is preserved if ``None``.

        file_op: see :obj:`snek.operations`, :obj:`~.create`` by default.

    .. versionchanged:: 4.0
       Instead of a ``update_rule`` flag, the function now accepts a :obj:`file_op
       <snek.oprtations.FileOp>`.

    Returns:
        Updated project tree representation

    Note:
        Use an empty string as content to ensure a file is created empty.
    """
    return modify(
        struct, path, lambda old, _: (old if content is None else content, file_op)
    )


def reject(struct: Structure, path: PathLike) -> Structure:
    """Remove a file from the project tree representation if existent.

    Args:
        struct: project representation as (possibly) nested.

        path: path-like string or object relative to the structure root.
            See :obj:`~.modify`.

            .. versionchanged:: 4.0
               The function no longer accepts a list of strings of path parts.

    Returns:
        Modified project tree representation
    """
    # Retrieve a list of parts from a path-like object
    path_parts = Path(path).parts

    # Walk the entire path, creating parents if necessary.
    root = deepcopy(struct)
    last_parent: dict = root
    name = path_parts[-1]
    for parent in path_parts[:-1]:
        if parent not in last_parent:
            return root  # one ancestor already does not exist, do nothing
        last_parent = last_parent[parent]

    if name in last_parent:
        del last_parent[name]

    return root


def merge(old: Structure, new: Structure) -> Structure:
    """Merge two dict representations for the directory structure.

    Basically a deep dictionary merge, except from the leaf update method.

    Args:
        old: directory descriptor that takes low precedence during the merge.
        new: directory descriptor that takes high precedence during the merge.


    .. versionchanged:: 4.0
        Project structure now considers everything **under** the
        top level project folder.

    Returns:
        Resulting merged directory representation

    Note:
        Use an empty string as content to ensure a file is created empty.
        (``None`` contents will not be created).
    """
    return _inplace_merge(deepcopy(old), new)


def _inplace_merge(old: Structure, new: Structure) -> Structure:
    """Similar to :obj:`~.merge` but modifies the first dict."""

    for key, value in new.items():
        old_value = old.get(key, None)
        new_is_dict = isinstance(value, dict)
        old_is_dict = isinstance(old_value, dict)
        if new_is_dict and old_is_dict:
            old[key] = _inplace_merge(
                cast(Structure, old_value), cast(Structure, value)
            )
        elif old_value is not None and not new_is_dict and not old_is_dict:
            # both are defined and final leaves
            old[key] = _merge_leaf(cast(Leaf, old_value), cast(Leaf, value))
        else:
            old[key] = deepcopy(value)

    return old


def _merge_leaf(old_value: Leaf, new_value: Leaf) -> Leaf:
    """Merge leaf values for the directory tree representation.

    The leaf value is expected to be a tuple ``(content, update_rule)``.
    When a string is passed, it is assumed to be the content and
    ``None`` is used for the update rule.

    Args:
        old_value: descriptor for the file that takes low precedence during the merge.
        new_value: descriptor for the file that takes high precedence during the merge.

    Note:
        ``None`` contents are ignored, use and empty string to force empty
        contents.

    Returns:
        Resulting value for the merged leaf
    """
    old = old_value if isinstance(old_value, (list, tuple)) else (old_value, None)
    new = new_value if isinstance(new_value, (list, tuple)) else (new_value, None)

    content = old[0] if new[0] is None else new[0]
    file_op = old[1] if new[1] is None else new[1]

    if file_op is None:
        return content

    return (content, file_op)



"""
Collection of functions responsible for the "reification" of the file representation in
the project structure into a file written to the disk.

A function that "reifies" a file (manifests the programmatic representation of the file
in the project structure dictionary to the disk) is called here a **file operation** or
simply a **file op**. Actually file ops don't even need to be functions strictly
speaking, they can be any :obj:`callable` object. The only restriction is that file ops
**MUST** respect the :obj:`FileOp` signature. The :obj:`create` function is a good
example of how to write a new file op.

A function (or callable) that modifies the behaviour of an existing file op, by wrapping
it with another function/callable is called here **modifier**. Modifiers work similarly
to Python `decorators`_ and allow extending/refining the existing file ops.
A modifier should receive a file op as argument and return another file op.
:obj:`no_overwrite` and :obj:`skip_on_update` are good examples on how to write new file
op modifiers.

While modifiers don't have a specific signature (the number of parameters might vary,
but they always return a single file op value), the following conventions **SHOULD** be
observed when creating new modifiers:

- Modifiers should accept at least one argument: the **file op** they modify
  (you don't have to be very creative, go ahead and name this parameter ``file_op``, it
  is a good convention).
  This parameter should be made **optional** (the default value of :obj:`create` usually
  works, unless there is a better default value for the main use case of the modifier).
- Modifiers can accept arguments other then the file op they modify. These arguments
  should **precede** the file op in the list of arguments (the file op should be the
  last argument, which interoperates well with :obj:`~functools.partial`).
- When writing a modifier that happens to be a function (instead of a callable class),
  please name the inner function with the same name of the modifier but preceded by an
  ``_`` (underscore) char. This allows better inspection/debugging.


.. versionchanged:: 4.0
   Previously, file operations where simply indicated as a numeric flag
   (the members of ``snek.structure.FileOp``) in the project structure.
   Starting from snek 4, file operation are functions with signature given by
   :obj:`FileOp`.

.. _decorators: https://en.wikipedia.org/wiki/Python_syntax_and_semantics#Decorators
.. _shebang: https://en.wikipedia.org/wiki/Shebang_(Unix)
.. _File access permissions: https://en.wikipedia.org/wiki/File_system_permissions
"""

from pathlib import Path
from typing import Any, Callable, Dict, Union

from . import file_system as fs
from .log import logger

# Signatures for the documentation purposes

ScaffoldOpts = Dict[str, Any]
"""Dictionary with snek's options, see :obj:`snek.api.create_project`.
Should be treated as immutable (if required, copy before changing).

Please notice some behaviours given by the options **SHOULD** be observed. For example,
files should be overwritten when the **force** option is ``True``. Similarly when
**pretend** is ``True``, no operation should be really performed, but any action should
be logged as if realized.
"""

FileContents = Union[str, None]
"""When the file content is ``None``, the file should not be written to
disk (empty files are represented by an empty string ``""`` as content).
"""

FileOp = Callable[[Path, FileContents, ScaffoldOpts], Union[Path, None]]
"""Signature of functions considered file operations::

    Callable[[Path, FileContents, ScaffoldOpts], Union[Path, None]]

Args:
    path (pathlib.Path): file path potentially to be written to/changed in the disk.
    contents (:obj:`FileContents`): usually a string that represents a text
        content of the file. :obj:`None` indicates the file should not be written.
    opts (:obj:`ScaffoldOpts`): a dict with snek's options.

Returns:
    If the file is written (or more generally changed, such as new access permissions),
    by convention they should return the :obj:`file path <pathlib.Path>`.
    If no file was touched, :obj:`None` should be returned. Please notice a **FileOp**
    might return :obj:`None` if a pre-existing file in the disk is not modified.

Note:
    A **FileOp** usually has side effects (e.g. write a file to the disk), see
    :obj:`FileContents` and :obj:`ScaffoldOpts` for other conventions.
"""


# FileOps and FileOp modifiers (a.k.a. factories/decorators/wrappers)


def create(path: Path, contents: FileContents, opts: ScaffoldOpts) -> Union[Path, None]:
    """
    Default :obj:`FileOp`: always create/write the file even during (forced) updates.
    """
    if contents is None:
        return None

    if not path.parent.is_dir():
        fs.create_directory(path.parent, pretend=opts.get("pretend"))

    return fs.create_file(path, contents, pretend=opts.get("pretend"))


def remove(path: Path, _content: FileContents, opts: ScaffoldOpts) -> Union[Path, None]:
    """Remove the file if it exists in the disk"""
    if not path.exists():
        return None

    return fs.rm_rf(path, pretend=opts.get("pretend"))


def no_overwrite(file_op: FileOp = create) -> FileOp:
    """File op modifier. Returns a :obj:`FileOp` that does not overwrite an existing
    file during update (still created if not exists).

    Args:
        file_op: a :obj:`FileOp` that will be "decorated",
            i.e. will be called if the ``no_overwrite`` conditions are met.
            Default: :obj:`create`.
    """

    def _no_overwrite(path: Path, contents: FileContents, opts: ScaffoldOpts):
        """See ``snek.operations.no_overwrite``"""
        if opts.get("force") or not path.exists():
            return file_op(path, contents, opts)

        logger.report("skip", path)
        return None

    return _no_overwrite


def skip_on_update(file_op: FileOp = create) -> FileOp:
    """File op modifier. Returns a :obj:`FileOp` that will skip the file during a
    project update (the file will just be created for brand new projects).

    Args:
        file_op: a :obj:`FileOp` that will be "decorated",
            i.e. will be called if the ``skip_on_update`` conditions are met.
            Default: :obj:`create`.
    """

    def _skip_on_update(path: Path, contents: FileContents, opts: ScaffoldOpts):
        """See ``snek.operations.skip_on_update``"""
        if opts.get("force") or not opts.get("update"):
            return file_op(path, contents, opts)

        logger.report("skip", path)
        return None

    return _skip_on_update


def add_permissions(permissions: int, file_op: FileOp = create) -> FileOp:
    """File op modifier. Returns a :obj:`FileOp` that will **add** access permissions to
    the file (on top of the ones given by default by the OS).

    Args:
        permissions (int): permissions to be added to file::

                updated file mode = old mode | permissions  (bitwise OR)

            Preferably the values should be a combination of
            :obj:`stat.S_* <stat.S_IRUSR>` values (see :obj:`os.chmod`).

        file_op: a :obj:`FileOp` that will be "decorated".
            If the file exists in disk after ``file_op`` is called (either created
            or pre-existing), ``permissions`` will be added to it.
            Default: :obj:`create`.

    Warning:
        This is an **experimental** file op and might be subject to incompatible changes
        (or complete removal) even in minor/patch releases.

    Note:
        `File access permissions`_ work in a completely different way depending on the
        operating system. This file op is straightforward on POSIX systems, but a bit
        tricky on Windows. It should be safe for desirable but not required effects
        (e.g. making a file with a `shebang`_ executable, but that can also run via
        ``python FILE``), or ensuring files are readable/writable.
    """

    def _add_permissions(path: Path, contents: FileContents, opts: ScaffoldOpts):
        """See ``snek.operations.add_permissions``"""
        return_value = file_op(path, contents, opts)

        if path.exists():
            mode = path.stat().st_mode | permissions
            return fs.chmod(path, mode, pretend=opts.get("pretend"))

        return return_value

    return _add_permissions



"""
Extension that omits the creation of file `cli.py`
"""

from pathlib import PurePath as Path
from typing import List

from .. import structure
from ..actions import Action, ActionParams, ScaffoldOpts, Structure
from . import Extension


class NoSkeleton(Extension):
    """Omit creation of cli.py and test_cli.py"""

    def activate(self, actions: List[Action]) -> List[Action]:
        """Activate extension, see :obj:`~snek.extension.Extension.activate`."""
        return self.register(actions, remove_files, after="define_structure")


def remove_files(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Remove all cli files from structure

    Args:
        struct: project representation as (possibly) nested :obj:`dict`.
        opts: given options, see :obj:`create_project` for an extensive list.

    Returns:
        Updated project representation and options
    """
    # Namespace is not yet applied so deleting from package is enough
    src = Path("src")
    struct = structure.reject(struct, src / opts["package"] / "cli.py")
    struct = structure.reject(struct, "tests/test_cli.py")
    return struct, opts



"""
Functionality to update one snek version to another
"""
from enum import Enum
from functools import reduce, wraps
from itertools import chain
from types import SimpleNamespace as Object
from typing import TYPE_CHECKING, Callable, Iterable, Tuple, cast

from configupdater import ConfigUpdater
from packaging.version import Version

from . import __version__ as snek_version
from . import dependencies as deps
from . import templates, toml
from .info import (
    PYPROJECT_TOML,
    SETUP_CFG,
    get_curr_version,
    read_pyproject,
    read_setupcfg,
)
from .log import logger
from .structure import ScaffoldOpts, Structure

if TYPE_CHECKING:  # pragma: no cover
    # ^  avoid circular dependencies in runtime
    from .actions import Action, ActionParams


(ALWAYS,) = list(Enum("VersionUpdate", "ALWAYS"))  # type: ignore
"""Perform the update action regardless of the version"""


def version_migration(struct: Structure, opts: ScaffoldOpts) -> "ActionParams":
    """Update projects that were generated with old versions of snek"""
    update = opts.get("update")

    if not update:
        return struct, opts

    from .actions import invoke  # delay import to avoid circular dependency error

    curr_version = get_curr_version(opts["project_path"])

    # specify how to migrate from one version to another as ordered list
    v4_plan = [
        replace_find_with_find_namespace,  # need to happen after update_setup_cfg
        handover_setup_requires,
        update_pyproject_toml,
    ]
    migration_plans = [
        (Version("3.1"), [add_entrypoints]),
        (ALWAYS, [update_setup_cfg, add_dependencies]),
        (Version("4.0"), v4_plan),
    ]

    plan_actions: Iterable["Action"] = chain.from_iterable(
        plan_actions
        for plan_version, plan_actions in migration_plans
        if plan_version is ALWAYS or curr_version < plan_version
    )

    # replace the old version with the updated one
    opts["version"] = snek_version
    return reduce(invoke, plan_actions, (struct, opts))


def _change_setupcfg(
    fn: Callable[[ConfigUpdater, ScaffoldOpts], Tuple[ConfigUpdater, ScaffoldOpts]]
) -> Callable[[Structure, ScaffoldOpts], "ActionParams"]:
    @wraps(fn)
    def _wrapped(struct: Structure, opts: ScaffoldOpts) -> "ActionParams":
        setupcfg = read_setupcfg(opts["project_path"])
        setupcfg, opts = fn(setupcfg, opts)
        if not opts["pretend"]:
            try:
                setupcfg.update_file()
            except Exception:  # pragma: no cover
                msg = f"Problems with {fn.__name__}. `setup.cfg` content:\n\n"
                logger.debug(msg + str(setupcfg) + "\n\n")
                raise

        logger.report("updated", opts["project_path"] / SETUP_CFG)
        return struct, opts

    return _wrapped


@_change_setupcfg
def add_entrypoints(setupcfg: ConfigUpdater, opts: ScaffoldOpts):
    """Add [options.entry_points] to setup.cfg"""
    new_section_name = "options.entry_points"
    if new_section_name in setupcfg:
        return setupcfg, opts

    cfg = ConfigUpdater().read_string(templates.setup_cfg(opts))
    new_section = cfg[new_section_name].detach()

    add_after_sect = "options.extras_require"
    if add_after_sect not in setupcfg:
        # user removed it for some reason, default to metadata
        add_after_sect = "metadata"

    setupcfg[add_after_sect].add_after.section(new_section).space()

    return setupcfg, opts


@_change_setupcfg
def update_setup_cfg(setupcfg: ConfigUpdater, opts: ScaffoldOpts):
    """Update `snek` in setupcfg and ensure some values are there as expected"""
    if "options" not in setupcfg:
        template = templates.setup_cfg(opts)
        new_section = ConfigUpdater().read_string(template)["options"]
        setupcfg["metadata"].add_after.section(new_section.detach())

    # Add "snek" section if missing and update saved extensions
    setupcfg = templates.add_snek(setupcfg, opts)
    return setupcfg, opts


@_change_setupcfg
def add_dependencies(setupcfg: ConfigUpdater, opts: ScaffoldOpts):
    """Add dependencies"""
    # TODO: Revise the need for `deps.RUNTIME` once `python_requires = >= 3.8`
    options = setupcfg["options"]
    if "install_requires" in options:
        install_requires = options.get("install_requires", Object(value=""))
        runtime_deps = deps.split(cast(str, install_requires.value))
        runtime_deps = deps.add(deps.RUNTIME, runtime_deps)
        options["install_requires"].set_values(runtime_deps)
    else:
        options.set("install_requires")
        options["install_requires"].set_values(deps.RUNTIME)

    return setupcfg, opts


@_change_setupcfg
def replace_find_with_find_namespace(setupcfg: ConfigUpdater, opts: ScaffoldOpts):
    setupcfg["options"].set("packages", "find_namespace:")
    return setupcfg, opts


# Ideally things involving ``no_pyproject`` should be implemented standalone in the
# NoPyProject extension... that is a bit hard though... So we take the pragmatic
# approach => implement things here (do nothing if the user is not using pyproject, but
# migrate the deprecated setup_requires over otherwise)


@_change_setupcfg
def handover_setup_requires(setupcfg: ConfigUpdater, opts: ScaffoldOpts):
    """When paired with :obj:`update_pyproject_toml`, this will transfer ``setup.cfg ::
    options.setup_requires`` to ``pyproject.toml :: build-system.requires``
    """
    options = setupcfg["options"]
    if "setup_requires" in options and opts.get("isolated_build", True):
        setup_requires = options.pop("setup_requires", Object(value=""))
        build_deps = deps.split(cast(str, setup_requires.value))
        opts.setdefault("build_deps", []).extend(build_deps)

    return setupcfg, opts


def update_pyproject_toml(struct: Structure, opts: ScaffoldOpts) -> "ActionParams":
    """Update old pyproject.toml generated by snekext-pyproject and import
    `setup_requires` from `update_setup_cfg` into `build-system.requires`.
    """

    if opts.get("pretend") or not opts.get("isolated_build", True):
        return struct, opts

    try:
        config = read_pyproject(opts["project_path"])
    except FileNotFoundError:
        # We still need to transfer ``setup_requires`` to pyproject.toml
        config = toml.loads(templates.pyproject_toml(opts))

    build = config["build-system"]
    existing = deps.add(opts.get("build_deps", []), build.get("requires", []))
    build["requires"] = deps.remove(deps.add(existing, deps.ISOLATED), ["snek"])
    # ^  snek is no longer a build dependency
    toml.setdefault(build, "build-backend", "setuptools.build_meta")
    toml.setdefault(config, "tool.setuptools_scm.version_scheme", "no-guess-dev")

    (opts["project_path"] / PYPROJECT_TOML).write_text(toml.dumps(config), "utf-8")
    logger.report("updated", opts["project_path"] / PYPROJECT_TOML)
    return struct, opts



import sys

import pytest

from snek import dependencies as deps


@pytest.mark.skipif(
    sys.version_info[:2] <= (3, 7),
    reason="regex doesn't work well in py36 (for some edge cases)",
)
def test_split():
    assert deps.split(
        "\n    snek>=42.1.0,<43.0"
        "\n    platformdirs==1"
        "\n    cookiecutter<8"
        "\n    mypkg~=9.0"
    ) == ["snek>=42.1.0,<43.0", "platformdirs==1", "cookiecutter<8", "mypkg~=9.0"]
    assert deps.split(
        "\n    snek>=42.1.0,<43.0;platformdirs==1"
        "\n    cookiecutter<8;mypkg~=9.0\n\n"
    ) == ["snek>=42.1.0,<43.0", "platformdirs==1", "cookiecutter<8", "mypkg~=9.0"]
    assert deps.split(
        "snek>=42.1.0,<43.0; platformdirs==1; cookiecutter<8; mypkg~=9.0; "
    ) == ["snek>=42.1.0,<43.0", "platformdirs==1", "cookiecutter<8", "mypkg~=9.0"]
    assert deps.split(
        "\n    snek>=42.1.0,<43.0;python_version>='3.4'; platformdirs==1"
    ) == ["snek>=42.1.0,<43.0;python_version>='3.4'", "platformdirs==1"]
    assert deps.split(
        "\n    snek>=42.1.0,<43.0; python_version>='3.4'; platformdirs==1"
    ) == ["snek>=42.1.0,<43.0; python_version>='3.4'", "platformdirs==1"]


def test_deduplicate():
    # no duplication => no effect
    assert deps.deduplicate(["snek>=4,<5", "platformdirs"]) == [
        "snek>=4,<5",
        "platformdirs",
    ]
    # duplicated => the last one wins
    assert deps.deduplicate(
        ["snek>=4,<5", "snek~=3.2", "snek==0"]
    ) == ["snek==0"]
    assert deps.deduplicate(
        ["snek==0", "snek>=4,<5", "snek~=3.2"]
    ) == ["snek~=3.2"]


def test_remove():
    assert deps.remove(
        ["snek>=42.1.0,<43.0", "platformdirs==1", "cookiecutter<8", "mypkg~=9.0"],
        ["platformdirs"],
    ) == ["snek>=42.1.0,<43.0", "cookiecutter<8", "mypkg~=9.0"]
    assert deps.remove(
        ["snek>=42.1.0,<43.0", "platformdirs==1", "cookiecutter<8", "mypkg~=9.0"],
        {"mypkg": 0},
    ) == ["snek>=42.1.0,<43.0", "platformdirs==1", "cookiecutter<8"]


def test_add():
    own_deps = [
        "setuptools_scm>=1.2.5,<2",
        "snek>=42.1.0,<43",
        "django>=5.3.99999,<6",
    ]
    # No intersection
    assert deps.add(["platformdirs==1", "cookiecutter<8", "mypkg~=9.0"], own_deps) == [
        "platformdirs==1",
        "cookiecutter<8",
        "mypkg~=9.0",
        "setuptools_scm>=1.2.5,<2",
        "snek>=42.1.0,<43",
        "django>=5.3.99999,<6",
    ]
    # With intersection => own_deps win
    assert deps.add(["platformdirs==1", "snek<8", "mypkg~=9.0"], own_deps) == [
        "platformdirs==1",
        "snek>=42.1.0,<43",
        "mypkg~=9.0",
        "setuptools_scm>=1.2.5,<2",
        "django>=5.3.99999,<6",
    ]


def test_add_commented():
    new_deps = [
        "setuptools_scm>=1.2.5,<2",
        "snek>=42.1.0,<43",
        "django>=5.3.99999,<6",
        "mypkg~=9.0",
        "gitdep @ git+https://repo.com/gitdep@main#egg=gitdep",
        "# a comment",
        "# comment-that>0==1<3; reminds.pep==508",
    ]
    assert deps.add(["platformdirs==1", "# a comment", "# mypkg~=2.0"], new_deps) == [
        "platformdirs==1",
        "# a comment",
        "mypkg~=9.0",
        "setuptools_scm>=1.2.5,<2",
        "snek>=42.1.0,<43",
        "django>=5.3.99999,<6",
        "gitdep @ git+https://repo.com/gitdep@main#egg=gitdep",
        "# comment-that>0==1<3; reminds.pep==508",
    ]



import logging
import re
import sys

import pytest

from snek.exceptions import ErrorLoadingExtension, exceptions2exit

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import EntryPoint  # pragma: no cover
else:
    from importlib_metadata import EntryPoint  # pragma: no cover


def test_exceptions2exit():
    @exceptions2exit([RuntimeError])
    def func(_):
        raise RuntimeError("Exception raised")

    with pytest.raises(SystemExit):
        func(1)


def test_exceptions2exit_verbose(capsys, monkeypatch):
    @exceptions2exit([RuntimeError])
    def func(_):
        raise RuntimeError("Exception raised")

    monkeypatch.setattr("snek.cli.get_log_level", lambda: logging.DEBUG)
    with pytest.raises(SystemExit):
        func(1)

    error = capsys.readouterr().err
    match = re.search(r"raise RuntimeError", error)
    assert match


def test_error_loading_external_extension():
    # Assert the error message displays some meaningful text
    extension = "snekext.fake.extension"

    # Extension name is given directly
    ex = str(ErrorLoadingExtension(extension))
    assert "an error loading" in ex
    assert "fake" in ex

    # Entrypoint is given
    fake = EntryPoint("fake", f"{extension}:Fake", "snek.cli")
    ex = str(ErrorLoadingExtension(entry_point=fake))
    assert "an error loading" in ex
    assert "fake" in ex



import logging
import os
import sys
from distutils.util import strtobool
from importlib import reload
from pathlib import Path
from tempfile import mkdtemp
from types import SimpleNamespace as Object

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib import metadata
else:
    import importlib_metadata as metadata

import pytest

from .helpers import (
    command_exception,
    disable_import,
    nop,
    replace_import,
    rmpath,
    uniqstr,
)
from .virtualenv import VirtualEnv

IS_POSIX = os.name == "posix"


def _config_git(home):
    config = """
        [user]
          name = Jane Doe
          email = janedoe@email
    """
    (home / ".gitconfig").write_text(config)


def _fake_expanduser(original_expand, real_home, fake_home):
    def _expand(path):
        value = original_expand(path)
        if value.startswith(str(fake_home)):
            return value

        return value.replace(str(real_home), str(fake_home))

    return _expand


@pytest.fixture(autouse=True)
def fake_home(tmp_path, monkeypatch):
    """Isolate tests.
    Avoid interference of an existing config dir in the developer's
    machine
    """
    real_home = os.getenv("REAL_HOME")
    home = os.getenv("HOME")
    if real_home and real_home != home:
        # Avoid doing it twice
        yield home
        return

    real_home = str(os.path.expanduser("~"))
    monkeypatch.setenv("REAL_HOME", real_home)

    fake = Path(mkdtemp(prefix="home", dir=str(tmp_path)))
    _config_git(fake)

    expanduser = _fake_expanduser(os.path.expanduser, real_home, fake)
    monkeypatch.setattr("os.path.expanduser", expanduser)
    monkeypatch.setenv("HOME", str(fake))
    monkeypatch.setenv("USERPROFILE", str(fake))  # Windows?

    yield fake
    rmpath(fake)


@pytest.fixture(autouse=True)
def fake_xdg_config_home(fake_home, monkeypatch):
    """Isolate tests.
    Avoid interference of an existing config dir in the developer's
    machine
    """
    home = str(fake_home)
    monkeypatch.setenv("XDG_CONFIG_HOME", home)
    yield home


@pytest.fixture(autouse=True)
def fake_config_dir(request, tmp_path, monkeypatch):
    """Isolate tests.
    Avoid interference of an existing config dir in the developer's
    machine
    """
    if "no_fake_config_dir" in request.keywords:
        # Some tests need to check the original implementation to make sure
        # side effects of the shared object are consistent. We have to try to
        # make them as few as possible.
        yield
        return

    confdir = Path(mkdtemp(prefix="conf", dir=str(tmp_path)))
    monkeypatch.setattr("snek.info.config_dir", lambda *_, **__: confdir)
    yield confdir
    rmpath(confdir)


@pytest.fixture
def venv(tmp_path, fake_home, fake_xdg_config_home):
    """Create a virtualenv for each test"""
    virtualenv = VirtualEnv(".venv", tmp_path)
    virtualenv.env["HOME"] = str(fake_home)
    virtualenv.env["USERPROFILE"] = str(fake_home)
    virtualenv.env["XDG_CONFIG_HOME"] = str(fake_xdg_config_home)

    trusted = os.environ.get("PIP_TRUSTED_HOST")
    if trusted:
        virtualenv.env["PIP_TRUSTED_HOST"] = trusted

    cache = os.environ.get("PIP_CACHE")
    if cache:
        virtualenv.env["PIP_CACHE"] = cache

    virtualenv.create()

    return virtualenv


@pytest.fixture
def existing_venv(venv):
    """Alias of ``venv`` to avoid clashes with ``snek.extensions.venv``"""
    return venv


@pytest.fixture
def snek():
    return __import__("snek")


@pytest.fixture
def real_isatty():
    snek = __import__("snek", globals(), locals(), ["termui"])
    return snek.termui.isatty


@pytest.fixture
def logger(monkeypatch):
    snek = __import__("snek", globals(), locals(), ["log"])
    logger_obj = snek.log.logger
    monkeypatch.setattr(logger_obj, "propagate", True)  # <- needed for caplog
    yield logger_obj


@pytest.fixture
def with_coverage():
    return strtobool(os.environ.get("COVERAGE", "NO"))


@pytest.fixture(autouse=True)
def isolated_logger(request, logger, monkeypatch):
    # In Python the common idiom of using logging is to share the same log
    # globally, even between threads. While this is usually OK because
    # internally Python takes care of locking the shared resources, it also
    # makes very difficult to build things on top of the logging system without
    # using the same global approach.
    # For simplicity, to make things easier to extension developers and because
    # Snek not really uses multiple threads, this is the case in
    # `snek.log`.
    # On the other hand, shared state and streams can make the testing
    # environment a real pain, since we are messing with everything all the
    # time, specially when running tests in parallel (so we not guarantee the
    # execution order).
    # This fixture do a huge effort in trying to isolate as much as possible
    # each test function regarding logging. We keep the global object, so the
    # tests can be seamless, but internally replace the underlying native
    # loggers and handlers for "one-shot" ones.
    # (Of course, we can keep the same global object just because the plugins
    # for running tests in parallel are based in multiple processes instead of
    # threads, otherwise we would need another strategy)

    if "original_logger" in request.keywords:
        # Some tests need to check the original implementation to make sure
        # side effects of the shared object are consistent. We have to try to
        # make them as few as possible.
        yield logger
        return

    # Get a fresh new logger, not used anywhere
    raw_logger = logging.getLogger(uniqstr())
    # ^  Python docs advert against instantiating Loggers directly and instruct
    #    devs to use `getLogger`. So we use a unique name to guarantee we get a
    #    new logger each time.
    raw_logger.setLevel(logging.NOTSET)
    new_handler = logging.StreamHandler()

    # Replace the internals of the LogAdapter
    # --> Messing with global state: don't try this at home ...
    #     (if we start to use threads, we cannot do this)

    # Be lazy to import modules due to coverage warnings
    # (see @FlorianWilhelm comments on #174)
    from snek.log import ReportFormatter

    monkeypatch.setattr(logger, "propagate", True)
    monkeypatch.setattr(logger, "nesting", 0)
    monkeypatch.setattr(logger, "wrapped", raw_logger)
    monkeypatch.setattr(logger, "handler", new_handler)
    monkeypatch.setattr(logger, "formatter", ReportFormatter())
    # <--

    try:
        yield logger
    finally:
        new_handler.close()
        # ^  Force the handler to not be re-used


@pytest.fixture
def tmpfolder(tmpdir):
    with tmpdir.as_cwd():
        yield tmpdir

    rmpath(tmpdir)


@pytest.fixture
def git_mock(monkeypatch, logger):
    def _git(*args, **kwargs):
        cmd = " ".join(["git"] + list(args))

        logger.report("run", cmd, context=os.getcwd())

        def _response():
            yield "git@mock"

        return _response()

    def _is_git_repo(folder):
        return Path(folder, ".git").is_dir()

    monkeypatch.setattr("snek.shell.git", _git)
    monkeypatch.setattr("snek.repo.is_git_repo", _is_git_repo)

    yield _git


@pytest.fixture
def nogit_mock(monkeypatch):
    def raise_error(*_):
        raise command_exception("No git mock!")

    monkeypatch.setattr("snek.shell.git", raise_error)
    yield


@pytest.fixture
def nogit_cmd_mock(monkeypatch):
    # With this fixture we still allow all the code paths in `get_git_cmd` to be
    # traversed during tests, so we improve the chances of catching errors.
    monkeypatch.setattr("snek.shell._GIT_CMD", "git-cmd.not-installed")
    monkeypatch.setattr("snek.shell._GIT_CMD_WIN", "git-cmd.not-installed.exe")

    from snek import shell

    shell.get_git_cmd.cache_clear()  # force reloading _GIT_CMD
    yield
    shell.get_git_cmd.cache_clear()  # force reloading _GIT_CMD


@pytest.fixture
def noconfgit_mock(monkeypatch):
    def raise_error(*argv):
        if "config" in argv:
            raise command_exception("No git mock!")

    monkeypatch.setattr("snek.shell.git", raise_error)
    yield


@pytest.fixture
def nodjango_admin_mock(monkeypatch):
    def raise_error(*_):
        raise command_exception("No django_admin mock!")

    monkeypatch.setattr("snek.shell.django_admin", raise_error)
    yield


@pytest.fixture
def nosphinx_mock():
    with disable_import("sphinx"):
        yield


@pytest.fixture
def version_raises_exception(monkeypatch, snek):
    def raise_exeception(name):
        raise metadata.PackageNotFoundError("No version mock")

    monkeypatch.setattr(metadata, "version", raise_exeception)
    reload(snek)
    try:
        yield
    finally:
        monkeypatch.undo()
        reload(snek)


@pytest.fixture(autouse=True)
def no_isatty(monkeypatch, real_isatty):
    # Requiring real_isatty ensures processing that fixture
    # before this one. Therefore real_isatty is cached before the mock
    # replaces the real function.

    # Avoid ansi codes in tests, since capture fixtures seems to
    # emulate stdout and stdin behavior (including isatty method)
    monkeypatch.setattr("snek.termui.isatty", lambda *_: False)
    yield


@pytest.fixture
def orig_isatty(monkeypatch, real_isatty):
    monkeypatch.setattr("snek.termui.isatty", real_isatty)
    yield real_isatty


@pytest.fixture
def no_curses_mock():
    with disable_import("curses"):
        yield


@pytest.fixture
def curses_mock():
    with replace_import("curses", Object()):
        yield


@pytest.fixture
def no_colorama_mock():
    with disable_import("colorama"):
        yield


@pytest.fixture
def colorama_mock():
    with replace_import("colorama", Object(init=nop)):
        yield



#!/usr/bin/env python
import sys
from pathlib import Path

from snek.api import create_project
from snek.cli import run
from snek.extensions import no_cli


def test_create_project_wit_no_cli(tmpfolder):
    # Given options with the no-cli extension,
    opts = dict(project_path="proj", extensions=[no_cli.NoSkeleton("no-cli")])

    # when the project is created,
    create_project(opts)

    # then cli file should not exist
    assert not Path("proj/src/proj/cli.py").exists()
    assert not Path("proj/tests/test_cli.py").exists()


def test_create_project_without_no_cli(tmpfolder):
    # Given options without the no-cli extension,
    opts = dict(project_path="proj")

    # when the project is created,
    create_project(opts)

    # then cli file should exist
    assert Path("proj/src/proj/cli.py").exists()
    assert Path("proj/tests/test_cli.py").exists()


def test_cli_with_no_cli(tmpfolder):
    # Given the command line with the no-cli option,
    sys.argv = ["snek", "--no-cli", "proj"]

    # when snek runs,
    run()

    # then cli file should not exist
    assert not Path("proj/src/proj/cli.py").exists()
    assert not Path("proj/tests/test_cli.py").exists()


def test_cli_with_no_cli_and_pretend(tmpfolder):
    # Given the command line with the no-cli and pretend options,
    sys.argv = ["snek", "--pretend", "--no-cli", "proj"]

    # when snek runs,
    run()

    # then cli file should not exist (or the project itself)
    assert not Path("proj/src/proj/cli.py").exists()
    assert not Path("proj").exists()


def test_cli_without_no_cli(tmpfolder):
    # Given the command line without the no-cli option,
    sys.argv = ["snek", "proj"]

    # when snek runs,
    run()

    # then cli file should exist
    assert Path("proj/src/proj/cli.py").exists()
    assert Path("proj/tests/test_cli.py").exists()



import logging
import os
import re
import shlex
import shutil
import stat
import sys
from pathlib import Path
from pprint import pformat

import pytest

from snek import shell

from .helpers import uniqstr


def test_ShellCommand(tmpfolder):
    echo = shell.ShellCommand("echo")
    output = echo("Hello Echo!!!")
    assert next(output).strip('"') == "Hello Echo!!!"
    python = shell.ShellCommand("python")
    output = python("-c", 'print("Hello World")')
    assert list(output)[-1] == "Hello World"
    touch = shell.ShellCommand("touch")
    touch("my-file.txt")
    assert Path("my-file.txt").exists()


def test_shell_command_error2exit_decorator():
    @shell.shell_command_error2exit_decorator
    def func(_):
        shell.ShellCommand("non_existing_cmd")("--wrong-args")

    with pytest.raises(SystemExit):
        func(1)


def test_command_exists():
    assert shell.command_exists("python")
    assert not shell.command_exists("ldfgyupmqzbch174")


def test_pretend_command(caplog):
    caplog.set_level(logging.INFO)
    # When command runs under pretend flag,
    name = uniqstr()
    touch = shell.ShellCommand("touch")
    touch(name, pretend=True)
    # then nothing should be executed
    assert not Path(name).exists()
    # but log should be displayed
    logs = caplog.text
    assert re.search(r"run.*touch\s" + name, logs)


def test_get_executable(tmpfolder):
    # Some python should exist
    assert shell.get_executable("python") is not None
    # No python should exist in an empty directory when the global $PATH is not included
    assert shell.get_executable("python", tmpfolder, include_path=False) is None
    # When using sys.prefix python should be sys.executable (+ version suffix)
    python = Path(sys.executable).resolve()
    bin_path = shell.get_executable("python", include_path=False, prefix=sys.prefix)
    bin_path = Path(bin_path).resolve()
    assert bin_path.stem in python.stem
    assert bin_path.parent == python.parent
    # Non existing binaries => None
    assert shell.get_executable(uniqstr()) is None


def test_get_command():
    python = shell.get_command("python", prefix=sys.prefix, include_path=False)
    assert next(python("--version")).strip().startswith("Python 3")
    with pytest.raises(shell.ShellCommandException):
        python("--" + uniqstr())


def test_get_command_inexistent():
    name = uniqstr()
    inexistent = shell.get_command(name, prefix=sys.prefix, include_path=False)
    assert inexistent is None


def test_get_command_with_whitespace(tmpfolder):
    # Given an executable exists in a path with spaces
    prefix = Path(tmpfolder, "with spaces")
    if os.name == "posix":
        executable = Path(prefix, "bin", "myexec")
        executable.parent.mkdir(parents=True, exist_ok=True)
        executable.write_text("#!/bin/sh\n\necho 42")
        executable.chmod(stat.S_IMODE(stat.S_IREAD | stat.S_IEXEC))
    elif os.name == "nt":  # Windows
        executable = Path(prefix, "Script", "myexec.bat")
        executable.parent.mkdir(parents=True, exist_ok=True)
        executable.write_text("@echo off\r\necho 42", encoding="ascii")
        # ^  Let's use a basic encoding + CRLF for windows
    else:
        pytest.skip("Requires either POSIX-compliant OS or Windows")
        return

    # ----> helps when debugging
    exec_path = shell.get_executable("myexec", prefix=prefix, include_path=False)
    print("exec_path:", pformat(shlex.quote(exec_path)))
    print("contents:\n", pformat(executable.read_text()))
    assert exec_path is not None
    assert Path(exec_path).exists()
    # <----

    # When we create a command with `get_command`
    cmd = shell.get_command("myexec", prefix=prefix, include_path=False)
    assert cmd is not None
    # it should run without any problems
    completed = cmd.run()
    print("stdout:", completed.stdout)
    assert int(completed.stdout) == 42
    completed.check_returncode()


def test_get_editor(monkeypatch):
    # In general we should always find an editor
    assert shell.get_editor() is not None
    # When there is a problem, then we should have a nice error message
    monkeypatch.delenv("VISUAL", raising=False)
    monkeypatch.setenv("EDITOR", "")
    monkeypatch.setattr(shell, "get_executable", lambda *_, **__: None)
    with pytest.raises(shell.ShellCommandException, match="set EDITOR"):
        print("editor", shell.get_editor())


def test_edit(tmpfolder, monkeypatch):
    vi = shutil.which("vim") or shutil.which("vi")
    if not vi:
        pytest.skip("This test requires `vim` or `vi` to be available")

    # Given a file exists
    file = tmpfolder / "test.txt"
    file.write_text("Hello World", "utf-8")
    assert file.read_text("utf-8").strip() == "Hello World"

    # Then `shell.edit` should be able to manipulate it
    monkeypatch.delenv("VISUAL", raising=False)
    monkeypatch.setenv("EDITOR", vi)
    shell.edit(file, "-c", ":%s/World/Snek/g", "-c", ":wq")
    # ^ a bit of vim scripting so it does not wait for the user to type

    assert file.read_text("utf-8").strip() == "Hello Snek"


def test_join():
    # Join should work with empty iterables
    assert shell.join([]) == ""
    assert shell.join({}) == ""
    assert shell.join(()) == ""
    assert shell.join(x for x in []) == ""

    # Join should accept Path objects
    p1 = Path("a", "b", "c")
    p2 = Path("d", "f", "g")
    assert shell.join([p1, p2]) == f"{p1} {p2}"

    # Join should be the opposite of split
    args = ["/my path/to exec", '"other args"', "asdf", "42", "'a b c'"]
    assert shlex.split(shell.join(args)) == args


def test_non_existent_file_exception():
    cmd = shell.ShellCommand("non_existent_cmd", shell=False)
    with pytest.raises(shell.ShellCommandException):
        cmd()



"""Similarly to ``git rebase -i`` this extension allows users to interactively
choose which options apply to ``putup``, by editing a file filled with examples.

See :obj:`CONFIG` for more details on how to tweak the text generated in the interactive
mode.

.. versionadded:: 4.0
   *"interactive mode"* introduced as an **experimental** extension.

Warning:

    **NOTE FOR CONTRIBUTORS**:
    Due to the way :mod:`argparse` is written, it is not very easy to obtain information
    about which options and arguments a given parser is currently configured with.
    There are no public methods that allow inspection/reflection, and therefore in order
    to do so, one has to rely on a few non-public methods (according to Python's
    convention, the ones starting with a ``_`` symbol).
    Since :mod:`argparse` implementation is very stable and mature, these non-public
    method are very unlikely to change and, therefore, it is relatively safe to use
    these methods, however developers and maintainers have to be aware and pay attention
    to eventual breaking changes.
    The non-public functions are encapsulated in the functions :obj:`get_actions` and
    :obj:`format_args` in this file, in order to centralise the usage of non-public API.
"""
import os
import shlex
import textwrap
from argparse import Action, ArgumentParser
from collections import abc
from functools import lru_cache, reduce
from itertools import chain
from typing import List, Optional, Set

from .. import api, cli, file_system, shell, templates
from ..actions import ScaffoldOpts as Opts
from ..actions import get_default_options
from . import Extension
from . import list_from_entry_points as list_all_extensions

INDENT_LEVEL = 4
HEADER = templates.get_template("header_interactive")


CONFIG_KEY = "interactive"
CONFIG = {
    "ignore": ["--help", "--version"],
    "comment": ["--verbose", "--very-verbose"],
}
"""Configuration for the options that are not associated with an extension class.
This dict value consist of a set of metadata organised as follows:

- Each value must be a list of "long" :obj:`argparse` option strings (e.g. `"--help"`
  instead of `"-h"`).
- Each key implies on a different interpretation for the metadata:

    - ``"ignore"``: Options that should be simply ignored when creating examples
    - ``"comment"``: Options that should be commented when creating examples,
      even if they appear in the original :obj:`sys.argv`.

Extension classes (or instances) can also provide configurations by defining a
``interactive`` attribute assigned to a similar :obj:`dict` object.
"""


@lru_cache(maxsize=2)
def get_config(kind: str) -> Set[str]:
    """Get configurations that will be used for generating examples
    (from both :obj:`CONFIG` and the ``interactive`` attribute of each extension).

    The ``kind`` argument can assume the same values as the :obj:`CONFIG` keys.

    This function is cached to improve performance. Call ``get_config.__wrapped__`` to
    bypass the cache (or ``get_config.cache_clear``, see :obj:`functools.lru_cache`).
    """
    # TODO: when `python_requires >= 3.8` use Literal["ignore", "comment"] instead of
    #       str for type annotation of kind
    configurable = CONFIG.keys()
    assert kind in configurable
    initial_value = set(CONFIG[kind])  # A set avoid repeated flags
    empty_config: dict = {k: [] for k in configurable}  # Same shape as CONFIG

    def _merge_config(accumulated_config, extension):
        # The main idea is to collect all the configuration dicts from the extensions
        # (`interactive` attribute) and merge them with CONFIG, avoiding repetitions
        extension_config_dict = getattr(extension, CONFIG_KEY, empty_config)
        return accumulated_config.union(set(extension_config_dict.get(kind, [])))

    return reduce(_merge_config, list_all_extensions(), initial_value)


class Interactive(Extension):
    """Interactively choose and configure snek's parameters"""

    parser: ArgumentParser

    def __init__(self, name: Optional[str] = None):
        super().__init__(name)
        setattr(self, CONFIG_KEY, {"ignore": [self.flag]})
        self.flags = [f"-{self.name[0].lower()}", self.flag]

    def augment_cli(self, parser: ArgumentParser):
        """See :obj:`~snek.extensions.Extension.augment_cli`."""
        self.parser = parser

        parser.add_argument(
            *self.flags,
            dest="command",
            action="store_const",
            const=self.command,
            help=self.help_text,
        )
        return self

    def command(self, opts: Opts):
        """This method replace the regular call to :obj:`cli.run_scaffold
        <snek.cli.run_scaffold>` with an intermediate file to confirm the user's
        choices in terms of arguments/options.
        """
        opts = expand_computed_opts(opts)
        examples = all_examples(self.parser, get_actions(self.parser), opts)
        content = (os.linesep * 2).join([HEADER.template, examples])
        with file_system.tmpfile(prefix="snek-", suffix=".args.sh") as file:
            file.write_text(content, "utf-8")
            content = shell.edit(file).read_text("utf-8")
            cli.main(split_args(content))  # Call the CLI again with the new options


def expand_computed_opts(opts: Opts) -> Opts:
    """Pre-process the given snek options and add default/computed values
    (including the ones derived from ``setup.cfg`` in case of ``--update`` or
    snek's own configuration file in the user's home directory.
    """
    _struct, opts = get_default_options({}, api.bootstrap_options(opts))
    return opts


def wrap(text: Optional[str], width=70, **kwargs) -> str:
    """Wrap text to fit lines with a maximum number of characters"""
    return os.linesep.join(textwrap.wrap(text or "", width, **kwargs))


def comment(text: str, comment_mark="#", indent_level=0):
    """Comment each line of the given text (optionally indenting it)"""
    return textwrap.indent(text, (" " * indent_level) + comment_mark + " ")


def join_block(*parts: str, sep=os.linesep):
    """Join blocks of text using ``sep``, but ignoring the empty ones"""
    return sep.join(p for p in parts if p)


def long_option(action: Action):
    """Get the long option corresponding to the given :obj:`argparse.Action`"""
    return sorted(action.option_strings or [""], key=len)[-1]


def alternative_flags(action: Action):
    """Get the alternative flags (i.e. not the long one) of a :obj:`argparse.Action`"""
    flags = sorted(action.option_strings, key=len)[:-1]
    return f"(or alternatively: {' '.join(flags)})" if flags else ""


def has_active_extension(action: Action, opts: Opts) -> bool:
    """Returns :obj:`True` if the given :obj:`argparse.Action` corresponds to an
    extension that was previously activated via CLI.
    """
    ext_flags = [getattr(ext, "flag", None) for ext in opts.get("extensions", [])]
    return any(f in ext_flags for f in action.option_strings)


def example_no_value(parser: ArgumentParser, action: Action, opts: Opts) -> str:
    """Generate a CLI example of option usage for a :obj:`argparse.Action` that do not
    expect arguments (``nargs = 0``).

    See :obj:`example`.
    """
    long = long_option(action)
    active_extension = has_active_extension(action, opts)
    value = opts.get(action.dest)
    stored_value = value == action.const and value != action.default
    # ^  This function is only invoked when `nargs == 0` (store_*)
    #    When the option is activated the value should be stored

    if long not in get_config("comment") and (active_extension or stored_value):
        return f" {long}"

    return comment(long)


def example_with_value(parser: ArgumentParser, action: Action, opts: Opts) -> str:
    """Generate a CLI example of option usage for a :obj:`argparse.Action` that expects
    one or more arguments (``nargs`` is ``"?"``, ``"*"``, ``"+"`` or ``"N" > 0``).

    See :obj:`example`.
    """
    long = long_option(action)
    arg = opts.get(action.dest)

    if action.nargs in [None, 1, "?"]:
        value = shlex.quote(f"{arg}")
    elif isinstance(arg, (abc.Sequence, abc.Set)):
        # We are expecting a sequence/set since nargs is *, + or N > 1
        value = " ".join(shlex.quote(f"{a}") for a in arg).strip()
    else:
        value = ""

    if arg is None or long in get_config("comment") or value == "":
        return comment(f"{long} {format_args(parser, action)}".strip())

    return f" {long} {value}"


def example(parser: ArgumentParser, action: Action, opts: Opts) -> str:
    """Generate a CLI example of option usage for the given :obj:`argparse.Action`.
    The ``opts`` argument corresponds to options already processed by snek, and
    interferes on the generated text (e.g., when the corresponding option is already
    processed, the example will be adjusted accordingly; when the
    corresponding option is not present, the example might be commented out; ...).

    This function will comment options that are marked in the ``"comment"``
    :obj:`configuration <CONFIG>`.
    """
    fn = example_no_value if action.nargs == 0 else example_with_value
    return fn(parser, action, opts)


def example_with_help(parser: ArgumentParser, action: Action, opts: Opts) -> str:
    """Generate a CLI example of option usage for the given :obj:`argparse.Action` that
    includes a comment text block explaining its meaning (basically the same text
    displayed when using ``--help``).

    See :obj:`example`.
    """
    return join_block(
        example(parser, action, opts),
        comment(alternative_flags(action), indent_level=INDENT_LEVEL),
        comment(wrap(action.help), indent_level=INDENT_LEVEL),
    )


def all_examples(parser: ArgumentParser, actions: List[Action], opts: Opts) -> str:
    """Generate a example of usage of the CLI options corresponding to the given
    :obj:`actions <argparse.Action>` including the help text.

    This function will skip options that are marked in the ``"ignore"``
    :obj:`configuration <CONFIG>`.

    See :obj:`example_with_help`.
    """
    parts = (
        example_with_help(parser, a, opts)
        for a in actions
        if long_option(a) not in get_config("ignore")
    )
    return join_block(*parts, sep=os.linesep * 3)


def split_args(text: str) -> List[str]:
    """Split the text from the interactive example into arguments that can be passed
    directly to :mod:`argparse`, as if they were invoked directly from the CLI
    (this includes removing comments).
    """
    lines = (line.strip() for line in text.splitlines())
    return list(chain.from_iterable(shlex.split(x) for x in lines if x and x[0] != "#"))


# -- Functions that encapsulate calls to argparse non-public API --


def get_actions(parser: ArgumentParser):
    """List actions related to options that were configured to the given
    :obj:`ArgumentParser`.

    Warning:
        This function uses non-public API from Python's stdlib :mod:`argparse`.
    """
    return parser._actions


def format_args(parser: ArgumentParser, action: Action) -> str:
    """Produce an example to be used together with the flag of the given action.

    Warning:
        This function uses non-public API from Python's stdlib :mod:`argparse`.
    """
    formatter = parser._get_formatter()
    return formatter._format_args(action, action.dest)



from os.path import isdir, isfile
from pathlib import Path

import pytest

from snek import actions, api, cli, operations, structure

NO_OVERWRITE = operations.no_overwrite()
SKIP_ON_UPDATE = operations.skip_on_update()


def test_create_structure(tmpfolder):
    struct = {
        "my_file": "Some content",
        "my_folder": {
            "my_dir_file": "Some other content",
            "empty_file": "",
            "file_not_created": None,
        },
        "empty_folder": {},
    }
    expected = {
        "my_file": "Some content",
        "my_folder": {"my_dir_file": "Some other content", "empty_file": ""},
        "empty_folder": {},
    }
    changed, _ = structure.create_structure(struct, {})

    assert changed == expected
    assert isdir("my_folder")
    assert isdir("empty_folder")
    assert isfile("my_folder/my_dir_file")
    assert isfile("my_folder/empty_file")
    assert not isfile("my_folder/file_not_created")
    assert isfile("my_file")
    assert open("my_file").read() == "Some content"
    assert open("my_folder/my_dir_file").read() == "Some other content"
    assert open("my_folder/empty_file").read() == ""


def test_create_structure_with_wrong_type(tmpfolder):
    with pytest.raises(TypeError):
        struct = {"strange_thing": 1}
        structure.create_structure(struct, {})


def test_create_structure_when_updating(tmpfolder):
    struct = {
        "my_file": "Some content",
        "my_folder": {"my_dir_file": "Some other content"},
        "empty_folder": {},
    }
    structure.create_structure(struct, dict(update=False))
    struct["my_folder"]["my_dir_file"] = "Changed content"
    structure.create_structure(struct, dict(update=True))
    with open("my_folder/my_dir_file") as fh:
        assert fh.read() == "Changed content"


def test_create_structure_create_project_folder(tmpfolder):
    struct = {"my_folder": {"my_dir_file": "Some other content"}}
    opts = dict(project_path="my_project", update=False)
    structure.create_structure(struct, opts)
    assert isdir("my_project")


def test_define_structure():
    args = ["project", "-p", "package", "-d", "description"]
    opts = cli.parse_args(args)
    opts = api.bootstrap_options(opts)
    _, opts = actions.get_default_options({}, opts)
    struct, _ = structure.define_structure({}, opts)
    assert isinstance(struct, dict)


def test_merge_basics():
    # Given an existing struct,
    struct = {"a": {"b": {"c": "1", "d": "2"}}}
    # when it is merged to another struct with some common folder
    extra_files = {"a": {"b": {"c": "0"}, "e": "2"}, "f": {"g": {"h": "0"}}}
    struct = structure.merge(struct, extra_files)
    # then the result, should contain both files from the original and the
    # merged struct,
    assert struct["a"]["b"]["d"] == "2"
    assert struct["f"]["g"]["h"] == "0"
    assert struct["a"]["e"] == "2"
    # the common leaves should be overridden and a tuple (content, rule)
    assert struct["a"]["b"]["c"] == "0"


def test_merge_rules_just_in_original():
    # When an update rule exists in the original struct,
    struct = {"a": {"b": ("0", SKIP_ON_UPDATE)}}
    # but not in the merged,
    extra_files = {"a": {"b": "3"}}
    struct = structure.merge(struct, extra_files)
    # then just the content should be updated
    # and the rule should be kept identical
    assert struct["a"]["b"] == ("3", SKIP_ON_UPDATE)


def test_merge_rules_just_in_merged():
    # When an update rule does not exist in the original struct,
    struct = {"a": {"b": "0"}}
    # but exists in the merged,
    extra_files = {"a": {"b": (None, SKIP_ON_UPDATE)}}
    struct = structure.merge(struct, extra_files)
    # then just the rule should be updated
    # and the content should be kept identical
    assert struct["a"]["b"] == ("0", SKIP_ON_UPDATE)


def test_empty_string_leads_to_empty_file_during_merge():
    # When the original struct contains a leaf,
    struct = {"a": {"b": "0"}}
    # and the merged struct overrides it with an empty content
    extra_files = {"a": {"b": ""}}
    struct = structure.merge(struct, extra_files)
    # then the resulting content should exist and be empty
    assert struct["a"]["b"] == ""


def test_modify_non_existent():
    # Given the original struct does not contain a leaf
    # that is targeted by the modify method,
    struct = {"a": {"b": "0"}}

    # When the modify is called
    # Then the argument passed should be None
    def _modifier(old, op):
        assert old is None
        return ("1", op)

    struct = structure.modify(struct, Path("a", "c"), _modifier)

    # But the result of the modifier function should be included in the tree
    assert struct["a"]["c"] == ("1", operations.create)


def test_modify_file_op():
    struct = {"a": {"b": "0"}}
    struct = structure.modify(struct, "a/b", lambda text, _: (text, SKIP_ON_UPDATE))
    assert struct["a"]["b"] == ("0", SKIP_ON_UPDATE)


def test_ensure_nested():
    # When the original struct does not contain a leaf
    struct = {"a": {"b": "0"}}
    # that is added using the ensure method,
    struct = structure.ensure(struct, Path("a", "c", "d", "e", "f"), content="1")
    # then all the necessary parent folder should be included
    assert isinstance(struct["a"]["c"], dict)
    assert isinstance(struct["a"]["c"]["d"], dict)
    assert isinstance(struct["a"]["c"]["d"]["e"], dict)
    # and the file itself
    assert struct["a"]["c"]["d"]["e"]["f"] == ("1", operations.create)


def test_ensure_overriden():
    # When the original struct contains a leaf
    struct = {"a": {"b": "0"}}
    # that is overridden using the ensure method,
    struct = structure.ensure(struct, Path("a", "b"), content="1")
    # and the file content should be overridden
    assert struct["a"]["b"] == ("1", operations.create)


def test_ensure_path():
    # When the ensure method is called with an string path
    struct = {}
    struct = structure.ensure(struct, "a/b/c/d", content="1")
    # Then the effect should be the same as if it were split
    assert struct["a"]["b"]["c"]["d"] == ("1", operations.create)


def test_ensure_file_op():
    # When the ensure method is called with a file_op and no content
    struct = {"a": {"b": "0"}}
    struct = structure.ensure(struct, "a/b", file_op=NO_OVERWRITE)
    # Then the content should remain the same
    # But the file_op should change
    assert struct["a"]["b"] == ("0", NO_OVERWRITE)


def test_reject():
    # When the original struct contain a leaf
    struct = {"a": {"b": {"c": "0"}}}
    # that is removed using the reject method,
    struct = structure.reject(struct, Path("a", "b", "c"))
    # then the struct should not contain the file
    assert "c" not in struct["a"]["b"]


def test_reject_without_ancestor():
    # Given a defined struct,
    struct = {"a": {"b": {"c": "0"}}}
    # when someone tries to remove a file using the reject method
    # but one of its ancestor does not exist in the struct,
    struct = structure.reject(struct, "a/b/x/c")
    # then the struct should be the same
    assert struct["a"]["b"]["c"] == "0"
    assert len(struct["a"]["b"]["c"]) == 1
    assert len(struct["a"]["b"]) == 1
    assert len(struct["a"]) == 1


def test_reject_without_file():
    # Given a defined struct,
    struct = {"a": {"b": {"c": "0"}}}
    # when someone tries to remove a file using the reject method
    # but one of its ancestor does not exist in the struct,
    struct = structure.reject(struct, "a/b/x")
    # then the struct should be the same
    assert struct["a"]["b"]["c"] == "0"
    assert len(struct["a"]["b"]["c"]) == 1
    assert len(struct["a"]["b"]) == 1
    assert len(struct["a"]) == 1



import snek


def test_version():
    version = snek.__version__.split(".")
    assert int(version[0]) >= 0


def test_unknown_version(version_raises_exception):
    version = snek.__version__
    assert version == "unknown"



"""
Default snek's actions and functions to manipulate them.

When generating a project, snek uses a pipeline of functions (each function will
receive as arguments the values returned by the previous function). These functions
have an specific purpose and are called **actions**. Please follow the :obj:`Action`
signature when developing your own action.

Note:
    Some actions are more complex and are placed in dedicated modules together with
    other auxiliary functions, see :mod:`snek.structure`,
    :mod:`snek.update`.
"""
import os
from datetime import date, datetime
from functools import reduce
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Tuple

from . import info, repo
from .exceptions import (
    ActionNotFound,
    DirectoryAlreadyExists,
    DirectoryDoesNotExist,
    GitDirtyWorkspace,
    InvalidIdentifier,
    NestedRepository,
)
from .identification import (
    deterministic_name,
    deterministic_sort,
    get_id,
    is_valid_identifier,
    make_valid_identifier,
)
from .log import logger
from .structure import Structure, create_structure, define_structure
from .update import version_migration

if TYPE_CHECKING:  # pragma: no cover
    from .extensions import Extension  # avoid circular dependencies in runtime

ScaffoldOpts = Dict[str, Any]
"""Dictionary with snek's options, see :obj:`snek.api.create_project`."""

Action = Callable[[Structure, ScaffoldOpts], Tuple[Structure, ScaffoldOpts]]
"""Signature of a snek action, both arguments should be treated as immutable,
but a copy of the arguments, modified by the extension might be returned::

    Callable[[Structure, ScaffoldOpts], Tuple[Structure, ScaffoldOpts]]

"""

ActionParams = Tuple[Structure, ScaffoldOpts]
"""Both argument and return type of an action ``(struct, opts)``,
so a sequence of actions work in pipeline::

    Tuple[Structure, ScaffoldOpts]

When actions run, they can return an updated copy of :obj:`Structure` and
:obj:`ScaffoldOpts`.
"""


# -------- Functions that deal with/manipulate actions --------


def discover(extensions: Iterable["Extension"]) -> List[Action]:
    """Retrieve the action list.

    This is done by concatenating the default list with the one generated after
    activating the extensions.

    Args:
        extensions: list of functions responsible for activating the extensions.
    """
    actions = DEFAULT.copy()
    extensions = deterministic_sort(extensions)

    # Activate the extensions
    actions = reduce(_activate, extensions, actions)

    # Deduplicate actions
    return list({deterministic_name(a): a for a in actions}.values())


def invoke(struct_and_opts: ActionParams, action: Action) -> ActionParams:
    """Invoke action with proper logging.

    Args:
        struct_and_opts: snek's arguments for actions
        action: to be invoked

    Returns:
        ActionParams: updated project representation and options
    """
    logger.report("invoke", get_id(action))
    with logger.indent():
        return action(*struct_and_opts)


def register(
    actions: List[Action],
    action: Action,
    before: Optional[str] = None,
    after: Optional[str] = None,
) -> List[Action]:
    """Register a new action to be performed during scaffold.

    Args:
        actions (List[Action]): previous action list.
        action (Action): function with two arguments: the first one is a
            (nested) dict representing the file structure of the project
            and the second is a dict with scaffold options.
            This function **MUST** return a tuple with two elements similar
            to its arguments. Example::

                def do_nothing(struct, opts):
                    return (struct, opts)

        **kwargs (dict): keyword arguments make it possible to choose a
            specific order when executing actions: when ``before`` or
            ``after`` keywords are provided, the argument value is used as
            a reference position for the new action. Example::

                register(actions, do_nothing,
                         after='create_structure')
                    # Look for the first action with a name
                    # `create_structure` and inserts `do_nothing` after it.
                    # If more than one registered action is named
                    # `create_structure`, the first one is selected.

                register(
                    actions, do_nothing,
                    before='snek.structure:create_structure')
                    # Similar to the previous example, but the probability
                    # of name conflict is decreased by including the module
                    # name.

            When no keyword argument is provided, the default execution
            order specifies that the action will be performed after the
            project structure is defined, but before it is written to the
            disk. Example::

                register(actions, do_nothing)
                    # The action will take place after
                    # `snek.structure:define_structure`

    Returns:
        List[Action]: modified action list.
    """
    reference = before or after or get_id(define_structure)
    position = _find(actions, reference)
    if not before:
        position += 1

    clone = actions[:]
    clone.insert(position, action)

    return clone


def unregister(actions: List[Action], reference: str) -> List[Action]:
    """Prevent a specific action to be executed during scaffold.

    Args:
        actions (List[Action]): previous action list.
        reference (str): action identifier. Similarly to the keyword
            arguments of :obj:`register` it can assume two formats:

                - the name of the function alone,
                - the name of the module followed by ``:`` and the name
                  of the function

    Returns:
        List[Action]: modified action list.
    """
    position = _find(actions, reference)
    return actions[:position] + actions[position + 1 :]


def _find(actions: Iterable[Action], name: str) -> int:
    """Find index of name in actions"""
    if ":" in name:
        names = [get_id(action) for action in actions]
    else:
        names = [action.__name__ for action in actions]

    try:
        return names.index(name)
    except ValueError:
        raise ActionNotFound(name)


# -------- snek's actions --------


def get_default_options(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Compute all the options that can be automatically derived.

    This function uses all the available information to generate sensible
    defaults. Several options that can be derived are computed when possible.

    Args:
        struct: project representation as (possibly) nested :obj:`dict`.
        opts: given options, see :obj:`create_project` for an extensive list.

    Returns:
        ActionParams: project representation and options with default values set

    Raises:
        :class:`~.DirectoryDoesNotExist`: when snek is told to
            update an nonexistent directory
        :class:`~.GitNotInstalled`: when git command is not available
        :class:`~.GitNotConfigured`: when git does not know user information

    Note:
        This function uses git to determine some options, such as author name
        and email.
    """
    # This function uses information from git, so make sure it is available
    info.check_git()

    project_path = str(opts.get("project_path", ".")).rstrip(os.sep)
    # ^  Strip (back)slash when added accidentally during update
    opts["project_path"] = Path(project_path)
    opts.setdefault("name", opts["project_path"].resolve().name)
    opts.setdefault("package", make_valid_identifier(opts["name"]))
    opts.setdefault("author", info.username())
    opts.setdefault("email", info.email())
    opts.setdefault("description", "")
    opts.setdefault("url", "")
    opts.setdefault("release_date", date.today().strftime("%Y-%m-%d"))
    # All kinds of derived parameters
    year = datetime.strptime(opts["release_date"], "%Y-%m-%d").year
    opts.setdefault("year", year)
    opts.setdefault(
        "title",
        "=" * len(opts["name"]) + "\n" + opts["name"] + "\n" + "=" * len(opts["name"]),
    )
    opts.setdefault("isolated_build", opts.setdefault("pyproject", True))

    # Initialize empty list of all requirements and extensions
    # (since not using deep_copy for the DEFAULT_OPTIONS, better add compound
    # values inside this function)
    opts.setdefault("requirements", [])
    opts.setdefault("extensions", [])
    opts.setdefault("root_pkg", opts["package"])
    opts.setdefault("qual_pkg", opts["package"])
    opts.setdefault("pretend", False)

    opts["license"] = info.best_fit_license(opts.get("license"))
    # ^ "Canonicalise" license

    return struct, opts


def verify_options_consistency(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Perform some sanity checks about the given options.

    Args:
        struct: project representation as (possibly) nested :obj:`dict`.
        opts: given options, see :obj:`create_project` for an extensive list.

    Returns:
        Updated project representation and options
    """
    if not is_valid_identifier(opts["package"]):
        raise InvalidIdentifier(
            f"Package name {opts['package']!r} is not a valid identifier."
        )

    if opts["update"] and not opts["force"]:
        if not info.is_git_workspace_clean(opts["project_path"]):
            raise GitDirtyWorkspace

    return struct, opts


def empty(path: Path):
    """Is the path empty?"""
    return not any(path.glob('*'))


def verify_project_dir(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Check if snek can materialize the project dir structure.

    Args:
        struct: project representation as (possibly) nested :obj:`dict`.
        opts: given options, see :obj:`create_project` for an extensive list.

    Returns:
        Updated project representation and options
    """
    project_path = opts["project_path"].resolve(strict=False)
    parent_path = project_path.parent
    logger.report("verify", f"does project path {project_path} exist...")
    if project_path.exists() and not empty(project_path):
        if not opts["update"] and not opts["force"]:
            raise DirectoryAlreadyExists(
                f"Directory {project_path} already exists! Use the `update` option to "
                "update an existing project or the `force` option to "
                "overwrite an existing directory."
            )
    elif opts["update"]:
        raise DirectoryDoesNotExist(
            f"Project {project_path} does not exist and thus cannot be updated!"
        )
    elif repo.is_git_repo(parent_path) and not opts["force"]:
        raise NestedRepository(parent_path)

    return struct, opts


def init_git(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Add revision control to the generated files.

    Args:
        struct: project representation as (possibly) nested :obj:`dict`.
        opts: given options, see :obj:`create_project` for an extensive list.

    Returns:
        Updated project representation and options
    """
    path = opts.get("project_path", ".")
    logger.report("check", f"is initialization of the git repository {path} needed...")
    if not opts["update"] and not repo.is_git_repo(path):
        repo.init_commit_repo(path, struct, pretend=opts.get("pretend"))

    return struct, opts


def report_done(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Just inform the user snek is done"""
    try:
        print("done! 🐍 🌟 ✨")
    except Exception:  # pragma: no cover
        print("done!")  # this exception is not really expected to happen
    return struct, opts


DEFAULT: List[Action] = [
    get_default_options,
    verify_options_consistency,
    define_structure,
    verify_project_dir,
    version_migration,
    create_structure,
    init_git,
    report_done,
]
"""Default list of actions forming the main pipeline executed by snek"""


# -------- Auxiliary functions --------


def _activate(actions: List[Action], extension: "Extension") -> List[Action]:
    """Activate extension with proper logging.
    The order of args is inverted to facilitate ``reduce``
    """
    logger.report("activate", extension.__module__)
    with logger.indent():
        return extension(actions)



from pathlib import Path

from snek.api import create_project
from snek.cli import run
from snek.extensions.no_pyproject import NoPyProject


def test_create_project_wit_no_pyproject(tmpfolder):
    # Given options with the no_pyproject extension,
    opts = dict(project_path="proj", extensions=[NoPyProject()])

    # when the project is created,
    _, opts = create_project(opts)

    # then file should not exist
    assert not Path("proj/pyproject.toml").exists()

    assert opts["isolated_build"] is False


def test_create_project_without_no_pyproject(tmpfolder):
    # Given options without the no_pyproject extension,
    opts = dict(project_path="proj")

    # when the project is created,
    _, opts = create_project(opts)

    # then file should exist
    assert Path("proj/pyproject.toml").exists()

    assert opts["isolated_build"] is True


def test_cli_with_no_pyproject(tmpfolder):
    # Given the command line with the no-pyproject option,
    # when snek runs,
    run(["--no-pyproject", "proj"])

    # then file should not exist
    assert not Path("proj/pyproject.toml").exists()


def test_cli_with_no_pyproject_and_pretend(tmpfolder):
    # Given the command line with the no-pyproject and pretend options,
    # when snek runs,
    run(["--pretend", "--no-pyproject", "proj"])

    # then file should not exist (or the project itself)
    assert not Path("proj/pyproject.toml").exists()
    assert not Path("proj").exists()


def test_cli_without_no_pyproject(tmpfolder):
    # Given the command line without the no-pyproject option,
    # when snek runs,
    run(["--no-pyproject", "proj"])

    # then file should not exist
    assert not Path("proj/pyproject.toml").exists()



import logging
import re
from os import getcwd
from os.path import abspath

import pytest

from snek.file_system import localize_path as lp
from snek.log import (
    DEFAULT_LOGGER,
    ColoredReportFormatter,
    ReportFormatter,
    ReportLogger,
    logger,
)

from .helpers import uniqstr
from .log_helpers import (
    ansi_pattern,
    ansi_regex,
    make_record,
    match_record,
    match_report,
)

# When adding tests to this file, please have in mind that the global shared
# logging strategy followed by Python and Snek can lead to pain while
# testing. Try to create a new ReportLogger with a brand new underlying native
# Logger object as much as possible (see test_pass_handler for an example), and
# just deactivate the `isolated_logger` fixture with the `original_logger` mark
# if really necessary.


@pytest.fixture
def uniq_raw_logger():
    return logging.getLogger(uniqstr())


# -- Handler Registration --


@pytest.mark.original_logger  # Avoid autouse fixture
def test_default_handler_registered():
    # When the module is imported,
    # Then a default handler should be registered.
    raw_logger = logging.getLogger(DEFAULT_LOGGER)
    assert raw_logger.handlers
    assert isinstance(raw_logger.handlers[0], logging.StreamHandler)
    assert isinstance(raw_logger.handlers[0].formatter, ReportFormatter)


def test_pass_handler(uniq_raw_logger):
    # When the report logger is created with a handler
    handler = logging.NullHandler()
    new_logger = ReportLogger(uniq_raw_logger, handler=handler)
    assert isinstance(new_logger.handler, logging.NullHandler)


# -- Formatter Registration --


@pytest.mark.original_logger
def test_default_formatter_registered():
    # When the module is imported,
    # Then a default formatter should be registered.
    raw_logger = logging.getLogger(DEFAULT_LOGGER)
    handler = raw_logger.handlers[0]
    assert isinstance(handler.formatter, ReportFormatter)


def test_pass_formatter(uniq_raw_logger):
    # When the report logger is created with a formatter
    # Then that formatter should be registered.
    formatter = logging.Formatter("%(levelname)s")
    new_logger = ReportLogger(uniq_raw_logger, formatter=formatter)
    assert new_logger.formatter == formatter


# -- LoggerAdapter --


@pytest.mark.original_logger
def test_log_adapter_respects_log_level(caplog):
    # Preferably, in order to change log levels in tests `caplog.set_level`
    # should be used. However, this is the one test where we use
    # `logging.getLogger(DEFAULT_LOGGER).setLevel` to test the
    # behavior reported in the documentation. The wrapper object
    # `snek.log.logger` should respect the log level if the underlying
    # python logger object have changed.
    caplog.set_level(logging.NOTSET)

    possible_levels = [
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]

    for level in possible_levels:
        assert logger.wrapped == logging.getLogger(DEFAULT_LOGGER)
        # Given the logging level is set to INFO using `logging.getLogger`
        logging.getLogger(DEFAULT_LOGGER).setLevel(level)
        # Then the logger level should change accordingly
        assert logger.level == level

    # Trying to not leave a big logging level configuration behind
    logging.getLogger(DEFAULT_LOGGER).setLevel(logging.NOTSET)


def test_report(caplog, tmpfolder):
    # Given the logger level is properly configured
    caplog.set_level(logging.INFO)
    # When the report method is called,
    name = uniqstr()
    logger.report("make", str(tmpfolder.join(name)))
    # Then the message should be formatted accordingly.
    logs = caplog.text
    match = re.search("make.+" + name, logs)
    assert match
    # And relative paths should be used
    assert lp("/tmp") not in match.group(0)


def test_indent(caplog):
    # Given the logger level is set to INFO,
    caplog.set_level(logging.INFO)
    lg = logger.copy()  # Create a local copy to avoid shared state
    # And the nesting level is known
    nesting = lg.nesting
    # When the report method is called within an indentation context,
    name = uniqstr()
    with lg.indent():
        lg.report("make", name)
    # Then the spacing should be increased.
    assert any(
        match_report(
            r,
            activity="make",
            content=name,
            spacing=ReportFormatter.SPACING * (nesting + 2),
        )
        for r in caplog.records
    )

    # When report is called within a multi level indentation context,
    count = 5
    name = uniqstr()
    with lg.indent(count):
        lg.report("make", name)
    # Then the spacing should be increased accordingly.
    assert any(
        match_report(
            r,
            activity="make",
            content=name,
            spacing=ReportFormatter.SPACING * (nesting + count + 1),
        )
        for r in caplog.records
    )

    # When any other method is called with indentation,
    count = 3
    name = uniqstr()
    with lg.indent(count):
        lg.info(name)
    # Then the spacing should be added in the beginning
    logs = caplog.text
    assert (ReportFormatter.SPACING * (nesting + count) + name) in logs


def test_copy(caplog):
    # Given the logger level is set to INFO,
    caplog.set_level(logging.INFO)
    lg = logger.copy()  # Create a local copy to avoid shared state
    # And the nesting level is known
    nesting = lg.nesting
    # And a copy of the logger is made within a context,
    count = 3
    with lg.indent(count):
        logger2 = lg.copy()
    # When the original logger indentation level is changed,
    name = uniqstr()
    with lg.indent(7):
        lg.report("make", "/some/report")
        # And the report method is called in the copy logger
        logger2.report("call", name)
        # Then the logging level should not be changed
        assert logger2.nesting == nesting + count

    # And the spacing should not be increased.
    assert any(
        match_report(
            r,
            activity="call",
            content=name,
            spacing=ReportFormatter.SPACING * (nesting + count + 1),
        )
        for r in caplog.records
    )


def test_reconfigure(monkeypatch, caplog, uniq_raw_logger):
    # Given an environment that supports color, and a restrictive logger
    caplog.set_level(logging.NOTSET)
    monkeypatch.setattr("snek.termui.supports_color", lambda *_: True)
    formatter = ReportFormatter()
    new_logger = ReportLogger(uniq_raw_logger, formatter=formatter, propagate=True)
    new_logger.level = logging.INFO
    # when the logger is reconfigured
    new_logger.reconfigure()
    name = uniqstr()
    # then the messages should be displayed and use colors
    new_logger.report("some1", name)
    out = caplog.messages[-1]
    assert re.search(ansi_pattern("some1") + ".+" + name, out)

    # when the logger is reconfigured with a higher level
    new_logger.reconfigure(log_level=logging.CRITICAL)
    # then the messages should not be displayed
    name = uniqstr()
    new_logger.report("some2", name)
    assert not re.search(ansi_pattern("some2") + ".+" + name, caplog.text)


def test_other_methods(caplog):
    # Given the logger level is properly set,
    caplog.set_level(logging.DEBUG)
    name = uniqstr()
    # When conventional methods are called on logger,
    logger.debug(name)
    # Then they should bypass `report`-specific formatting
    assert all(not match_report(r, message=name) for r in caplog.records)
    assert any(
        match_record(r, levelno=logging.DEBUG, message=name) for r in caplog.records
    )


# -- Formatter --


def test_create_padding():
    formatter = ReportFormatter()
    for text in ["abcd", "abcdefg", "ab"]:
        padding = formatter.create_padding(text)
        # Formatter should ensure activates are right padded
        assert len(padding + text) == formatter.ACTIVITY_MAXLEN


def parent_dir():
    return abspath(lp(".."))


def test_format_path():
    formatter = ReportFormatter()
    format = formatter.format_path
    # Formatter should abbrev paths but keep other subjects unchanged
    assert format("not a command") == "not a command"
    assert format("git commit") == "git commit"
    assert format("a random message") == "a random message"
    assert format(getcwd()) == "."
    assert format(lp("../dir/../dir/..")) == lp("..")
    assert format(lp("../dir/../dir/../foo")) == lp("../foo")
    # shorter absolute is better than relative
    assert format(lp("/a")) == lp("/a")


def test_format_target():
    formatter = ReportFormatter()
    format = formatter.format_target
    assert format(None) == ""
    assert format(getcwd()) == ""
    assert format(parent_dir()) == "to '..'"


def test_format_context():
    formatter = ReportFormatter()
    format = formatter.format_context
    assert format(None) == ""
    assert format(getcwd()) == ""
    assert format(parent_dir()) == "from '..'"


def test_format():
    formatter = ReportFormatter()

    def format(*args, **kwargs):
        return formatter.format(make_record(*args, **kwargs)).lstrip()

    assert format("run", "ls -lf .") == "run  ls -lf ."
    assert format("run", "ls", context=parent_dir()) == "run  ls from '..'"
    assert format(
        "copy", getcwd(), target=lp("../dir/../dir")
    ) == "copy  . to '{}'".format(lp("../dir"))
    fmt_out = format("create", lp("my/file"), nesting=1)
    assert fmt_out == f"create    {lp('my/file')}"


def test_colored_format_target():
    formatter = ColoredReportFormatter()
    format = formatter.format_target
    out = format(parent_dir())
    assert ColoredReportFormatter.TARGET_PREFIX in out
    assert ansi_regex("to").search(out)


def test_colored_format_context():
    formatter = ColoredReportFormatter()
    format = formatter.format_context
    out = format(parent_dir())
    assert ColoredReportFormatter.CONTEXT_PREFIX in out
    assert ansi_regex("from").search(out)


def test_colored_activity():
    formatter = ColoredReportFormatter()
    format = formatter.format_activity
    out = format("run")
    assert ansi_regex("run").search(out)


def test_colored_format():
    formatter = ColoredReportFormatter()

    def format(*args, **kwargs):
        return formatter.format(make_record(*args, **kwargs)).lstrip()

    out = format("invoke", "action")
    assert ansi_regex("invoke").search(out)
    assert ansi_regex("action").search(out)


def test_colored_report(tmpfolder, caplog, uniq_raw_logger):
    # Given the logger is properly set,
    uniq_raw_logger.setLevel(logging.INFO)
    formatter = ColoredReportFormatter()
    uniq_logger = ReportLogger(uniq_raw_logger, formatter=formatter, propagate=True)
    # When the report method is called,
    name = uniqstr()
    uniq_logger.report("make", str(tmpfolder.join(name)))
    # Then the message should contain activity surrounded by ansi codes,
    out = caplog.messages[-1]
    assert re.search(ansi_pattern("make") + ".+" + name, out)
    # And relative paths should be used
    assert lp("/tmp") not in out


def test_colored_others_methods(caplog, uniq_raw_logger):
    # Given the logger is properly set,
    uniq_raw_logger.setLevel(logging.DEBUG)
    formatter = ColoredReportFormatter()
    uniq_logger = ReportLogger(uniq_raw_logger, formatter=formatter, propagate=True)
    # When conventional methods are called on logger,
    name = uniqstr()
    uniq_logger.debug(name)
    # Then the message should be surrounded by ansi codes
    out = caplog.messages[-1]
    assert ansi_regex(name).search(out)



"""
Functions for exception manipulation + custom exceptions used by Snek to identify
common deviations from the expected behavior.
"""
import functools
import logging
import sys
import traceback
from pathlib import Path
from typing import Optional, Sequence, Union, cast

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import EntryPoint  # pragma: no cover
else:
    from importlib_metadata import EntryPoint  # pragma: no cover

from . import __version__ as snek_version


def exceptions2exit(exception_list):
    """Decorator to convert given exceptions to exit messages

    This avoids displaying nasty stack traces to end-users

    Args:
        exception_list [Exception]: list of exceptions to convert
    """

    def exceptions2exit_decorator(func):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except tuple(exception_list) as ex:  # noqa: B030
                from .cli import get_log_level  # defer circular imports to avoid errors

                if get_log_level() <= logging.DEBUG:
                    # user surely wants to see the stacktrace
                    traceback.print_exc()
                print(f"ERROR: {ex}")
                sys.exit(1)

        return func_wrapper

    return exceptions2exit_decorator


class ActionNotFound(KeyError):
    """Impossible to find the required action."""

    def __init__(self, name, *args, **kwargs):
        message = ActionNotFound.__doc__[:-1] + f": `{name}`"
        super().__init__(message, *args, **kwargs)


class DirectErrorForUser(RuntimeError):
    """Parent class for exceptions that can be shown directly as error messages
    to the final users.
    """

    def __init__(self, message=None, *args, **kwargs):
        message = message or self.__class__.__doc__
        super().__init__(message, *args, **kwargs)


class DirectoryAlreadyExists(DirectErrorForUser):
    """The project directory already exists, but no ``update`` or ``force``
    option was used.
    """


class DirectoryDoesNotExist(DirectErrorForUser):
    """No directory was found to be updated."""


class GitNotInstalled(DirectErrorForUser):
    """Snek requires git to run."""

    DEFAULT_MESSAGE = (
        "Make sure git is installed and working. "
        "Use flag --very-verbose for more information."
    )

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class GitNotConfigured(DirectErrorForUser):
    """Snek tries to read user.name and user.email from git config."""

    DEFAULT_MESSAGE = (
        "Make sure git is configured. Run:\n"
        '  git config --global user.email "you@example.com"\n'
        '  git config --global user.name "Your Name"\n'
        "to set your account's default identity."
    )

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class GitDirtyWorkspace(DirectErrorForUser):
    """Workspace of git is empty."""

    DEFAULT_MESSAGE = (
        "Your working tree is dirty. Commit your changes first" " or use '--force'."
    )

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class InvalidIdentifier(RuntimeError):
    """Python requires a specific format for its identifiers.

    https://docs.python.org/3.6/reference/lexical_analysis.html#identifiers
    """


class SnekTooOld(DirectErrorForUser):
    """Snek cannot update a pre 3.0 version"""

    DEFAULT_MESSAGE = (
        "setup.cfg has no section [snek]! "
        "Are you trying to update a pre 3.0 version?"
    )

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class NoSnekProject(DirectErrorForUser):
    """Snek cannot update a project that it hasn't generated"""

    DEFAULT_MESSAGE = "Could not update project. Was it generated with Snek?"

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class ShellCommandException(RuntimeError):
    """Outputs proper logging when a ShellCommand fails"""


class ImpossibleToFindConfigDir(DirectErrorForUser):
    """An expected error occurred when trying to find the config dir.

    This might be related to not being able to read the $HOME env var in Unix
    systems, or %USERPROFILE% in Windows, or even the username.
    """


class ExtensionNotFound(DirectErrorForUser):
    """The following extensions were not found: {extensions}.
    Please make sure you have the required versions installed.
    """

    def __init__(self, extensions: Sequence[str]):
        message = cast(str, self.__doc__)
        message = message.format(extensions=extensions, version=snek_version)
        super().__init__(message)


class ErrorLoadingExtension(DirectErrorForUser):
    """There was an error loading '{extension}'.
    Please make sure you have installed a version of the extension that is compatible
    with Snek {version}. You can also try unininstalling it.
    """

    def __init__(self, extension: str = "", entry_point: Optional[EntryPoint] = None):
        if entry_point and not extension:
            extension = getattr(entry_point, "module", entry_point.name)

        if extension.endswith(".extension"):
            extension = extension[: -len(".extension")]
        extension = extension.replace("snekext.", "snekext-")

        message = cast(str, self.__doc__)
        message = message.format(extension=extension, version=snek_version)
        super().__init__(message)


class NestedRepository(DirectErrorForUser):
    """The directory '{directory}' is already part of a git repository.

    Snek avoids creating nested projects to prevent errors with
    `setuptools-scm`.

    If you know what you are doing you can try running `putup` again with
    the`--force` flag, but please be aware that you will have to manually
    customise the configuration for `setuptools-scm`. For more information,
    have a look on:

    - https://github.com/pypa/setuptools_scm
    """

    def __init__(self, directory: Union[str, Path]):
        message = cast(str, self.__doc__)
        super().__init__(message.format(directory=directory))



import logging
import os
import shlex
import sys
import traceback
from importlib.util import find_spec
from os import environ
from pathlib import Path
from shutil import which
from subprocess import STDOUT, CalledProcessError, check_output
from typing import List, Tuple

import pytest

IS_POSIX = os.name == "posix"

PYTHON = sys.executable
"""Same python executable executing the tests... Hopefully the one inside the virtualenv
inside tox folder. If we install packages by mistake is not a huge problem.
"""


def find_package_bin(package, binary=None):
    """If a ``package`` can be executed via ``python -m`` (with the current python)
    try to do that, otherwise use ``binary`` on the $PATH"""
    binary = binary or package
    if find_spec(package):
        return f"{PYTHON} -m {package}"

    executable = which(binary)
    if executable:
        msg = "Package %s can not be found inside %s, using system executable %s"
        logging.critical(msg, package, sys.prefix, executable)
        return executable

    pytest.skip(f"For some reason {binary} cannot be found.")


def find_venv_bin(venv_path, bin_name):
    """Given the path for a venv, find a executable there"""
    assert Path(venv_path).exists()
    candidates = list(map(str, Path(venv_path).glob(f"*/{bin_name}*")))
    return sorted(candidates, key=len)[0]


def merge_env(other=None, **kwargs):
    """Create a dict from merging items to the current ``os.environ``"""
    env = {k: v for k, v in environ.items()}  # Clone the environ as a dict
    env.update(other or {})
    env.update(kwargs)
    return env


def normalize_run_args(args: List[str], kwargs: dict) -> Tuple[List[str], dict]:
    # normalize args
    if len(args) == 1:
        if isinstance(args[0], str):
            args = shlex.split(args[0], posix=IS_POSIX)

    opts = dict(stderr=STDOUT, universal_newlines=True, encoding="utf-8")
    opts.update(kwargs)

    return [str(x) for x in args], opts


def run_with_debug(args: List[str], kwargs: dict):
    """Run the external command. See ``subprocess.check_output``."""
    try:
        return check_output(args if len(args) > 1 else args[0], **kwargs).strip()
    except CalledProcessError as ex:
        print("Error while running command:")
        print(args)
        print(kwargs)
        traceback.print_exc()
        msg = "******************** Terminal ($? = {}) ********************\n{}"
        print(msg.format(ex.returncode, ex.output))
        raise


def run(*args, **kwargs):
    """Run the external command. See ``subprocess.check_output``."""
    args, kwargs = normalize_run_args(args, kwargs)
    if args[0] in ("python", "putup", "pip", "tox", "pytest", "pre-commit"):
        raise SystemError("Please specify an executable with explicit path")
    return run_with_debug(args, kwargs)


def sphinx_cmd(build):
    docs_dir = Path("docs")
    build_dir = docs_dir / "_build"
    doctrees = build_dir / "doctrees"
    output_dir = build_dir / build
    return f"{PYTHON} -m sphinx -b {build} -d {doctrees} {docs_dir} {output_dir}"


def run_common_tasks(tests=True, flake8=True):
    if tests:
        run(f"{PYTHON} -m pytest", env=merge_env(PYTHONPATH="src"))

    run(sphinx_cmd("doctest"))
    run(sphinx_cmd("html"))

    run(f"{PYTHON} -m build .")

    if flake8 and environ.get("COVERAGE") == "true":
        run(f"{PYTHON} -m flake8 --count")



"""
Extension that adjust project file tree to include a namespace package.

This extension adds a **namespace** option to
:obj:`~snek.api.create_project` and provides correct values for the
options **root_pkg** and **namespace_pkg** to the following functions in the
action list.
"""

import os
from pathlib import Path
from typing import List, cast

from ..actions import Action, ActionParams, ScaffoldOpts, Structure
from ..exceptions import InvalidIdentifier
from ..file_system import chdir, move
from ..identification import is_valid_identifier
from ..log import logger
from ..operations import remove
from . import Extension, store_with


class Namespace(Extension):
    """Add a namespace (container package) to the generated package."""

    def augment_cli(self, parser):
        """Add an option to parser that enables the namespace extension.

        Args:
            parser (argparse.ArgumentParser): CLI parser object
        """
        parser.add_argument(
            self.flag,
            dest=self.name,
            default=None,
            action=store_with(self),
            metavar="NS1[.NS2]",
            help="put your project inside a namespace package "
            "(default: use no namespace)",
        )
        return self

    def activate(self, actions: List[Action]) -> List[Action]:
        """Register an action responsible for adding namespace to the package.

        Args:
            actions: list of actions to perform

        Returns:
            list: updated list of actions
        """
        actions = self.register(
            actions, enforce_namespace_options, after="get_default_options"
        )

        actions = self.register(actions, add_namespace, before="version_migration")

        return self.register(actions, move_old_package, after="create_structure")


def prepare_namespace(namespace_str: str) -> List[str]:
    """Check the validity of namespace_str and split it up into a list

    Args:
        namespace_str: namespace, e.g. "com.blue_yonder"

    Returns:
        list of namespaces, e.g. ["com", "com.blue_yonder"]

    Raises:
          :obj:`InvalidIdentifier` : raised if namespace is not valid
    """
    namespaces = namespace_str.split(".") if namespace_str else list()
    for namespace in namespaces:
        if not is_valid_identifier(namespace):
            raise InvalidIdentifier(f"{namespace} is not a valid namespace package.")
    return [".".join(namespaces[: i + 1]) for i in range(len(namespaces))]


def enforce_namespace_options(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Make sure options reflect the namespace usage."""
    opts.setdefault("namespace", "")

    if opts["namespace"]:
        opts["ns_list"] = prepare_namespace(opts["namespace"])
        opts["root_pkg"] = opts["ns_list"][0]
        opts["qual_pkg"] = ".".join([opts["ns_list"][-1], opts["package"]])
        opts["namespace"] = opts["namespace"].strip()

    return struct, opts


def add_namespace(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Prepend the namespace to a given file structure

    Args:
        struct: directory structure as dictionary of dictionaries
        opts: options of the project

    Returns:
        Directory structure as dictionary of dictionaries and input options
    """
    if not opts.get("namespace"):
        msg = "Using the `Namespace` extension with an empty namespace string/None. "
        msg += "You can try a valid string with the `--namespace` option in `putup` "
        msg += "(snek CLI) the arguments in `create_project` (snek API)."
        logger.warning(msg)
        return struct, opts

    namespace = opts["ns_list"][-1].split(".")
    base_struct = struct
    struct = cast(Structure, base_struct["src"])  # recursive types not supported yet
    pkg_struct = cast(Structure, struct[opts["package"]])
    del struct[opts["package"]]
    for sub_package in namespace:
        struct[sub_package] = {"__init__.py": ("", remove)}  # convert to PEP420
        struct = cast(Structure, struct[sub_package])
    struct[opts["package"]] = pkg_struct

    return base_struct, opts


def move_old_package(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Move old package that may be eventually created without namespace

    Args:
        struct (dict): directory structure as dictionary of dictionaries
        opts (dict): options of the project

    Returns:
        tuple(dict, dict):
            directory structure as dictionary of dictionaries and input options
    """
    project_path = Path(opts.get("project_path", "."))
    with chdir(project_path, **opts):
        old_path = Path("src", opts["package"])
        namespace_path = opts["qual_pkg"].replace(".", os.sep)
        target = Path("src", namespace_path)

        old_exists = opts["pretend"] or old_path.is_dir()
        #  ^  When pretending, pretend also an old folder exists
        #     to show a worst case scenario log to the user...

        if old_exists and opts["qual_pkg"] != opts["package"]:
            if not opts["pretend"]:
                logger.warning(
                    "\nA folder %r exists in the project directory, and it "
                    "is likely to have been generated by a snek "
                    "extension or manually by one of the current project "
                    "authors.\n"
                    "Moving it to %r, since a namespace option was passed.\n"
                    "Please make sure to edit all the files that depend on  "
                    "this package to ensure the correct location.\n",
                    opts["package"],
                    namespace_path,
                )

            move(old_path, target=target, pretend=opts["pretend"])

    return struct, opts



import os
import stat
from unittest.mock import Mock

from snek.operations import (
    add_permissions,
    create,
    no_overwrite,
    remove,
    skip_on_update,
)

from .helpers import temp_umask, uniqpath


def test_create(monkeypatch):
    created = {}

    def spy(path, *_1, **_2):
        created.update({path: True})

    monkeypatch.setattr("snek.file_system.create_file", spy)

    # When content is string, execute, even if it is empty
    for contents in ("contents", ""):
        path = uniqpath()
        create(path, contents, {})
        assert created[path]

    # When content is None, skip
    path = uniqpath()
    create(path, None, {})
    assert path not in created


def test_remove(monkeypatch):
    removed = {}

    def spy(path, *_1, **_2):
        removed.update({path: True})

    monkeypatch.setattr("pathlib.Path.exists", lambda _: True)
    monkeypatch.setattr("snek.file_system.rm_rf", spy)

    for contents in ("contents", "", None):
        path = uniqpath()
        remove(path, contents, {})
        assert removed[path]

    # Non existing paths are simply skipped
    monkeypatch.setattr("pathlib.Path.exists", lambda _: False)
    path = uniqpath()
    assert remove(path, contents, {}) is None
    assert path not in removed


def test_skip_on_update(monkeypatch):
    SKIP_ON_UPDATE = skip_on_update(lambda *_: path)

    # When update is False, execute
    opts = {"update": False}
    path = uniqpath()
    assert SKIP_ON_UPDATE(path, "contents", opts) == path

    # When update is True, skip, even if file does not exist
    opts = {"update": True}
    path = uniqpath()
    for existing in (True, False):
        with monkeypatch.context() as patch:
            patch.setattr("pathlib.Path.exists", Mock(return_value=existing))
            assert SKIP_ON_UPDATE(path, "contents", opts) is None

    # When force is True, execute, even if file exists
    opts = {"update": True, "force": True}
    path = uniqpath()
    for existing in (True, False):
        with monkeypatch.context() as patch:
            patch.setattr("pathlib.Path.exists", Mock(return_value=existing))
            assert SKIP_ON_UPDATE(path, "contents", opts) == path


def test_no_overwrite(monkeypatch):
    NO_OVERWRITE = no_overwrite(lambda *_: path)

    # When file does not exist, execute
    path = uniqpath()
    with monkeypatch.context() as patch:
        patch.setattr("pathlib.Path.exists", lambda _: False)
        assert NO_OVERWRITE(path, "contents", {}) == path

    # When file exists, skip
    path = uniqpath()
    with monkeypatch.context() as patch:
        patch.setattr("pathlib.Path.exists", lambda _: True)
        assert NO_OVERWRITE(path, "contents", {}) is None

    # When force is True, execute, even if file exists
    opts = {"force": True}
    path = uniqpath()
    for existing in (True, False):
        with monkeypatch.context() as patch:
            patch.setattr("pathlib.Path.exists", Mock(return_value=existing))
            assert NO_OVERWRITE(path, "contents", opts) == path


def test_add_permissions(tmpfolder):
    _ = tmpfolder  # Just used for chdir

    with temp_umask(0o333):
        # Without add_permissions, mode should be the same as umask
        path = uniqpath()
        create(path, "contents", {})
        assert stat.S_IMODE(path.stat().st_mode) == 0o444

        # With add_permissions, mode should have the extra bits
        path = uniqpath()
        if os.name == "posix":
            add_permissions(stat.S_IXOTH)(path, "contents", {})
            # ^  executable permissions should be the most common use case
            assert stat.S_IMODE(path.stat().st_mode) == 0o445
        else:
            add_permissions(stat.S_IWRITE)(path, "contents", {})
            # ^  windows executables work in a complete different way, so we just do a
            #    basic test with writeable access, just for the sake of completeness
            assert stat.S_IMODE(path.stat().st_mode) == 0o666



"""
All template files (`*.template`) within this subpackage are licensed
under the BSD 0-Clause license.
"""

import os
import string
import sys
from types import ModuleType
from types import SimpleNamespace as Object
from typing import Any, Dict, Set, Union, cast

from configupdater import ConfigUpdater

from .. import __version__ as snek_version
from .. import dependencies as deps
from .. import toml

if sys.version_info[:2] >= (3, 7):
    # TODO: Import directly (no need for workaround) when `python_requires = >= 3.7`
    from importlib.resources import read_text  # pragma: no cover
else:
    from pkgutil import get_data  # pragma: no cover

    def read_text(package, resource, encoding="utf-8") -> str:  # pragma: no cover
        data = get_data(package, resource)
        if data is None:
            raise FileNotFoundError(f"{resource!r} resource not found in {package!r}")
        return data.decode(encoding)


ScaffoldOpts = Dict[str, Any]

#: All available licences (identifiers based on SPDX ``https://spdx.org/licenses/``)
licenses = {
    "MIT": "license_mit",
    "AGPL-3.0-or-later": "license_affero_3.0",
    "AGPL-3.0-only": "license_affero_3.0",
    "Apache-2.0": "license_apache",
    "Artistic-2.0": "license_artistic_2.0",
    "0BSD": "license_bsd0",
    "BSD-2-Clause": "license_simplified_bsd",
    "BSD-3-Clause": "license_new_bsd",
    "CC0-1.0": "license_cc0_1.0",
    "EPL-1.0": "license_eclipse_1.0",
    "GPL-2.0-or-later": "license_gpl_2.0",
    "GPL-2.0-only": "license_gpl_2.0",
    "GPL-3.0-or-later": "license_gpl_3.0",
    "GPL-3.0-only": "license_gpl_3.0",
    "ISC": "license_isc",
    "LGPL-2.0-or-later": "license_lgpl_2.1",
    "LGPL-2.0-only": "license_lgpl_2.1",
    "LGPL-3.0-or-later": "license_lgpl_3.0",
    "LGPL-3.0-only": "license_lgpl_3.0",
    "MPL-2.0": "license_mozilla",
    "Unlicense": "license_public_domain",
    # ---- not really in the SPDX license list ----
    "Proprietary": "license_none",
}
# order is relevant: -only licenses should come after -or-later, so they dominate
# MIT goes first so it behaves like the default if an empty string is passed


def get_template(
    name: str, relative_to: Union[str, ModuleType] = __name__
) -> string.Template:
    """Retrieve the template by name

    Args:
        name: name of template (the ``.template`` extension will be
            automatically added to this name)
        relative_to: module/package object or name to which the resource file
            is relative (in the standard module format, e.g. ``foo.bar.baz``).
            Notice that ``relative_to`` should not represent directly a shared
            namespace package, since this kind of package is spread in
            different folders in the file system.

            Default value: ``snek.templates``
            (**please assign accordingly when using in custom extensions**).

    Examples:
        Consider the following package organization::

            .
            ├── src
            │   └── my_package
            │       ├── __init__.py
            │       ├── templates
            │       │   ├── __init__.py
            │       │   ├── file1.template
            │       │   └── file2.template
            │       …
            └── tests

        Inside the file ``src/my_package/__init__.py``, one can easily obtain
        the contents of ``file1.template`` by doing:

        .. code-block:: python

            from snek.templates import get_template
            from . import templates as my_templates

            tpl1 = get_template("file1", relative_to=my_templates)
            # OR
            # tpl1 = get_template('file1', relative_to=my_templates.__name__)

    Please notice you can also use `relative_to=__name__`
    or a combination of `from .. import __name__ as parent` and
    `relative_to=parent` to deal with relative imports.

    Returns:
        :obj:`string.Template`: template

    .. versionchanged :: 3.3
        New parameter **relative_to**.
    """
    file_name = f"{name}.template"
    if isinstance(relative_to, ModuleType):
        relative_to = relative_to.__name__

    data = read_text(relative_to, file_name, encoding="utf-8")
    # we assure that line endings are converted to '\n' for all OS
    content = data.replace(os.linesep, "\n")
    return string.Template(content)


def setup_cfg(opts: ScaffoldOpts) -> str:
    """Template of setup.cfg

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """

    template = get_template("setup_cfg")

    # template needs single-line `description`,
    # thus we pass only the first line of a multi-line description...
    desc = opts.get("description", "").splitlines() or [""]
    cfg_str = template.substitute({**opts, "description": desc[0]})

    updater = ConfigUpdater()
    updater.read_string(cfg_str)

    # ... and finally the multi-line string for the full description.
    if desc[0] != opts.get("description", ""):
        updater["metadata"]["description"].set_values(desc)

    requirements = deps.add(deps.RUNTIME, opts.get("requirements", []))
    updater["options"]["install_requires"].set_values(requirements)

    # fill [snek] section used for later updates
    add_snek(updater, opts)
    snek = updater["snek"]
    snek["version"].add_after.option("package", opts["package"])

    return str(updater)


def add_snek(config: ConfigUpdater, opts: ScaffoldOpts) -> ConfigUpdater:
    """Add snek section to a ``setup.cfg``-like file + snek's version +
    extensions and their associated options.
    """
    if "snek" not in config:
        config.add_section("snek")

    snek = config["snek"]
    snek["version"] = snek_version

    # Add the new extensions alongside the existing ones
    extensions = {ext.name for ext in opts.get("extensions", []) if ext.persist}
    old = cast(str, snek.get("extensions", Object(value="")).value)
    new = list(sorted(parse_extensions(old) | extensions))
    if new:
        snek.set("extensions")
        snek["extensions"].set_values(new)

    # Add extension-related opts, i.e. opts which start with an extension name
    allowed = ((k, v) for k, v in opts.items() if any(map(k.startswith, extensions)))
    allowed_ = {k: "" if v is None else v for k, v in allowed}
    # ^  TODO: remove workaround for None values once ConfigUpdater solves #68
    snek.update(allowed_)

    return config


def parse_extensions(extensions: str) -> Set[str]:
    """Given a string value for the field ``snek.extensions`` in a
    ``setup.cfg``-like file, return a set of extension names.
    """
    ext_names = (ext.strip() for ext in extensions.strip().split("\n"))
    return {ext for ext in ext_names if ext}


def pyproject_toml(opts: ScaffoldOpts) -> str:
    template = get_template("pyproject_toml")
    config = toml.loads(template.safe_substitute(opts))
    config["build-system"]["requires"] = list(deps.ISOLATED)
    return toml.dumps(config)


def license(opts):
    """Template of LICENSE.txt

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template(licenses[opts["license"]])
    return template.substitute(opts)


def init(opts):
    """Template of __init__.py

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    if opts["package"] == opts["name"]:
        opts["distribution"] = "__name__"
    else:
        opts["distribution"] = '"{}"'.format(opts["name"])
    template = get_template("__init__")
    return template.substitute(opts)



"""Extension that generates configuration for Cirrus CI."""
from argparse import ArgumentParser
from typing import List

from .. import structure
from ..actions import Action, ActionParams, ScaffoldOpts, Structure
from ..operations import no_overwrite
from ..templates import get_template
from . import Extension, include
from .pre_commit import PreCommit

TEMPLATE_FILE = "cirrus"


class Cirrus(Extension):
    """Add configuration file for Cirrus CI (includes `--pre-commit`)"""

    def augment_cli(self, parser: ArgumentParser):
        """Augments the command-line interface parser.
        See :obj:`~snek.extension.Extension.augment_cli`.
        """
        parser.add_argument(
            self.flag, help=self.help_text, nargs=0, action=include(PreCommit(), self)
        )
        return self

    def activate(self, actions: List[Action]) -> List[Action]:
        """Activate extension, see :obj:`~snek.extension.Extension.activate`."""
        return self.register(actions, add_files, after="define_structure")


def add_files(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Add .cirrus.yml to the file structure

    Args:
        struct: project representation as (possibly) nested :obj:`dict`.
        opts: given options, see :obj:`create_project` for an extensive list.

    Returns:
        struct, opts: updated project representation and options
    """
    files: Structure = {".cirrus.yml": (cirrus_descriptor, no_overwrite())}

    return structure.merge(struct, files), opts


def cirrus_descriptor(_opts: ScaffoldOpts) -> str:
    """Returns the rendered template"""
    return get_template(TEMPLATE_FILE).template  # no substitutions required



"""Internal library that encapsulate file system manipulation.
Examples include: creating/removing files and directories, changing permissions, etc.

Functions in this library usually extend the behaviour of Python's standard lib by
providing proper error handling or adequate logging/control flow in the context of
snek (an example of adequate control flow logic is dealing with the ``pretend``
flag).
"""

import errno
import os
import shutil
import stat
from contextlib import contextmanager
from functools import partial
from pathlib import Path
from tempfile import mkstemp
from typing import Callable, Optional, Union

from .log import logger
from .shell import IS_WINDOWS

PathLike = Union[str, os.PathLike]


@contextmanager
def tmpfile(**kwargs):
    """Context manager that yields a temporary :obj:`Path`"""
    fp, path = mkstemp(**kwargs)
    os.close(fp)  # we don't need the low level file handler
    file = Path(path)
    try:
        yield file
    finally:
        file.unlink()


@contextmanager
def chdir(path: PathLike, **kwargs):
    """Contextmanager to change into a directory

    Args:
        path : path to change current working directory to

    Keyword Args:
        pretend (bool): skip execution (but log) when pretending.
            Default ``False``.
    """
    should_pretend = kwargs.get("pretend")
    # ^ When pretending, automatically output logs
    #   (after all, this is the primary purpose of pretending)

    curr_dir = os.getcwd()

    try:
        logger.report("chdir", path)
        with logger.indent():
            if not should_pretend:
                os.chdir(path)
            yield
    finally:
        os.chdir(curr_dir)


def move(*src: PathLike, target: PathLike, **kwargs):
    """Move files or directories to (into) a new location

    Args:
        *src (PathLike): one or more files/directories to be moved

    Keyword Args:
        target (PathLike): if target is a directory, ``src`` will be
            moved inside it. Otherwise, it will be the new path (note that it
            may be overwritten)
        pretend (bool): skip execution (but log) when pretending.
            Default ``False``.
    """
    should_pretend = kwargs.get("pretend")

    for path in src:
        if not should_pretend:
            shutil.move(str(path), str(target))
        logger.report("move", path, target=target)


def create_file(path: PathLike, content: str, pretend=False, encoding="utf-8"):
    """Create a file in the given path.

    This function reports the operation in the logs.

    Args:
        path: path in the file system where contents will be written.
        content: what will be written.
        pretend (bool): false by default. File is not written when pretending,
            but operation is logged.

    Returns:
        Path: given path
    """
    path = Path(path)
    if not pretend:
        path.write_text(content, encoding=encoding)

    logger.report("create", path)
    return path


def create_directory(path: PathLike, update=False, pretend=False) -> Optional[Path]:
    """Create a directory in the given path.

    This function reports the operation in the logs.

    Args:
        path: path in the file system where contents will be written.
        update (bool): false by default. A :obj:`OSError` can be raised
            when update is false and the directory already exists.
        pretend (bool): false by default. Directory is not created when
            pretending, but operation is logged.
    """
    path = Path(path)
    if path.is_dir() and update:
        logger.report("skip", path)
        return None

    if not pretend:
        try:
            path.mkdir(parents=True, exist_ok=True)
        except OSError:
            if not update:
                raise
            return path  # Do not log if not created

    logger.report("create", path)
    return path


def chmod(path: PathLike, mode: int, pretend=False) -> Path:
    """Change the permissions of file in the given path.

    This function reports the operation in the logs.

    Args:
        path: path in the file system whose permissions will be changed
        mode: new permissions, should be a combination of
            :obj`stat.S_* <stat.S_IXUSR>` (see :obj:`os.chmod`).
        pretend (bool): false by default. File is not changed when pretending,
            but operation is logged.
    """
    path = Path(path)
    mode = stat.S_IMODE(mode)

    if not pretend:
        path.chmod(mode)

    logger.report(f"chmod {mode:03o}", path)
    return path


def localize_path(path_string: str) -> str:
    """Localize path for Windows, Unix, i.e. / or \\

    Args:
        path_string (str): path using /

    Returns:
        str: path depending on OS
    """
    return str(Path(path_string))


#: Windows-specific error code indicating an invalid pathname.
ERROR_INVALID_NAME = 123


def is_pathname_valid(pathname: str) -> bool:
    """Check if a pathname is valid

    Code by Cecil Curry from StackOverflow

    Args:
        pathname (str): string to validate

    Returns:
        `True` if the passed pathname is a valid pathname for the current OS;
        `False` otherwise.
    """
    # If this pathname is either not a string or is but is empty, this pathname
    # is invalid.
    try:
        if not isinstance(pathname, str) or not pathname:
            return False

        # Strip this pathname's Windows-specific drive specifier (e.g., `C:\`)
        # if any. Since Windows prohibits path components from containing `:`
        # characters, failing to strip this `:`-suffixed prefix would
        # erroneously invalidate all valid absolute Windows pathnames.
        _, pathname = os.path.splitdrive(pathname)

        # Directory guaranteed to exist. If the current OS is Windows, this is
        # the drive to which Windows was installed (e.g., the "%HOMEDRIVE%"
        # environment variable); else, the typical root directory.
        root_dirname = os.environ.get("HOMEDRIVE", "C:") if IS_WINDOWS else os.path.sep
        assert os.path.isdir(root_dirname)  # ...Murphy and her ironclad Law

        # Append a path separator to this directory if needed.
        root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

        # Test whether each path component split from this pathname is valid or
        # not, ignoring non-existent and non-readable path components.
        for pathname_part in pathname.split(os.path.sep):
            try:
                os.lstat(root_dirname + pathname_part)
            # If an OS-specific exception is raised, its error code
            # indicates whether this pathname is valid or not. Unless this
            # is the case, this exception implies an ignorable kernel or
            # filesystem complaint (e.g., path not found or inaccessible).
            #
            # Only the following exceptions indicate invalid pathnames:
            #
            # * Instances of the Windows-specific "WindowsError" class
            #   defining the "winerror" attribute whose value is
            #   "ERROR_INVALID_NAME". Under Windows, "winerror" is more
            #   fine-grained and hence useful than the generic "errno"
            #   attribute. When a too-long pathname is passed, for example,
            #   "errno" is "ENOENT" (i.e., no such file or directory) rather
            #   than "ENAMETOOLONG" (i.e., file name too long).
            # * Instances of the cross-platform "OSError" class defining the
            #   generic "errno" attribute whose value is either:
            #   * Under most POSIX-compatible OSes, "ENAMETOOLONG".
            #   * Under some edge-case OSes (e.g., SunOS, *BSD), "ERANGE".
            except OSError as exc:
                if hasattr(exc, "winerror"):
                    if exc.winerror == ERROR_INVALID_NAME:  # type: ignore
                        return False
                elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                    return False
    # If a "TypeError" exception was raised, it almost certainly has the
    # error message "embedded NUL character" indicating an invalid pathname.
    except TypeError:
        return False
    # If no exception was raised, all path components and hence this
    # pathname itself are valid. (Praise be to the curmudgeonly python.)
    else:
        return True
    # If any other exception was raised, this is an unrelated fatal issue
    # (e.g., a bug). Permit this exception to unwind the call stack.
    #
    # Did we mention this should be shipped with Python already?


def on_ro_error(func, path, exc_info):
    """Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.

    Usage : ``shutil.rmtree(path, onerror=onerror)``

    Args:
        func (callable): function which raised the exception
        path (str): path passed to `func`
        exc_info (tuple of str): exception info returned by sys.exc_info()
    """
    import stat
    from time import sleep

    # Sometimes the SO is just asynchronously (??!) slow, but it does remove the file
    sleep(0.5)

    if not Path(path).exists():
        return

    if not os.access(path, os.W_OK):
        # Is the error an access error ?
        os.chmod(path, stat.S_IWUSR)
        return func(path)

    raise


def rm_rf(path: PathLike, pretend=False):
    """Remove ``path`` by all means like ``rm -rf`` in Linux"""
    target = Path(path)
    if not target.exists():
        return None

    if target.is_dir():
        remove: Callable = partial(shutil.rmtree, onerror=on_ro_error)
    else:
        remove = Path.unlink

    if not pretend:
        remove(target)

    logger.report("remove", target)
    return path



import sys
from configparser import ConfigParser
from pathlib import Path

import pytest

from snek import actions, api
from snek import dependencies as deps
from snek import info, templates


def test_get_template():
    template = templates.get_template("setup_py")
    content = template.safe_substitute()
    assert content.split("\n", 1)[0] == '"""'


@pytest.fixture
def tmp_python_path(tmp_path):
    sys.path.append(str(tmp_path))
    yield tmp_path
    sys.path.remove(str(tmp_path))


def test_get_template_relative_to(tmp_python_path):
    # Given a template exists inside a package
    parent = tmp_python_path / "pkg4test"
    pkg = tmp_python_path / "pkg4test" / "asdf42_123456"
    pkg.mkdir(parents=True, exist_ok=True)
    (parent / "__init__.py").touch(exist_ok=True)
    (parent / "ex1.template").write_text("${var1}")
    (pkg / "__init__.py").touch(exist_ok=True)
    (pkg / "ex2.template").write_text("${var2}")

    # When using "relative_to" with __name__
    import pkg4test

    tpl1 = templates.get_template("ex1", relative_to=pkg4test.__name__)
    content = tpl1.safe_substitute({"var1": "Hello World!"})
    # Then get_template should work
    assert content == "Hello World!"

    # When using "relative_to" with a module
    import pkg4test

    tpl1 = templates.get_template("ex1", relative_to=pkg4test)
    content = tpl1.safe_substitute({"var1": "Some World!"})
    # Then get_template should work
    assert content == "Some World!"

    # When using "relative_to" with a package name string
    tpl2 = templates.get_template("ex2", relative_to="pkg4test.asdf42_123456")
    content = tpl2.safe_substitute({"var2": "Bye bye World!"})
    # Then get_template should work
    assert content == "Bye bye World!"


def test_all_licenses():
    opts = {
        "email": "test@user",
        "name": "my_project",
        "author": "myself",
        "year": 1832,
    }
    for license in templates.licenses.keys():
        opts["license"] = license
        assert templates.license(opts)


def test_setup_cfg():
    reqs = ("mydep1>=789.8.1", "mydep3<=90009;python_version>'3.5'", "other")
    opts = api.bootstrap_options({"project_path": "myproj", "requirements": reqs})
    _, opts = actions.get_default_options({}, opts)
    text = templates.setup_cfg(opts)
    setup_cfg = ConfigParser()
    setup_cfg.read_string(text)

    # Assert install_requires is correctly assigned
    install_requires = deps.split(setup_cfg["options"]["install_requires"])
    for dep in reqs:
        assert dep in install_requires
    # Assert Snek section
    assert setup_cfg["snek"].get("version")


def test_setup_cfg_2line_description(tmpfolder):
    # When a 2 line description is found (e.g. by reading an existing setup.cfg file)
    _, opts = actions.get_default_options({}, {"project_path": tmpfolder})
    opts["description"] = "2 line\ndescription"
    # Then the rendered template should still be valid
    text = templates.setup_cfg(opts)
    setup_cfg = ConfigParser()
    setup_cfg.read_string(text)
    assert setup_cfg["metadata"]["description"].strip() == "2 line\ndescription"

    Path(tmpfolder, "setup.cfg").write_text(text)
    opts = info.project({})
    assert opts["description"].strip() == "2 line\ndescription"



"""Runner module for demoapp"""

import argparse
import sys

import demoapp


def parse_args(args):
    """Parse command line parameters

    Args:
        args: command line parameters as list of strings

    Returns:
        :obj:`argparse.Namespace`: command line parameters
    """
    parser = argparse.ArgumentParser(
        description="A demo application for Snek's unit testing"
    )
    version = demoapp.__version__
    parser.add_argument(
        "-v", "--version", action="version", version=f"demoapp {version}"
    )
    opts = parser.parse_args(args)
    return opts


def main(args):
    parse_args(args)
    print("Hello World")


def run():
    """
    Entry point for setup.py
    """
    main(sys.argv[1:])


if __name__ == "__main__":
    main(sys.argv[1:])



import json
import os
import sys
from functools import partial
from pathlib import Path
from subprocess import CalledProcessError

import pytest

from snek import __version__ as snek_version
from snek.api import create_project
from snek.extensions.venv import Venv

from .helpers import find_venv_bin, run

pytestmark = [pytest.mark.slow, pytest.mark.system]


@pytest.fixture(autouse=True)
def dont_load_dotenv(monkeypatch):
    """pytest-virtualenv creates a `.env` directory by default, but `.env`
    entries in the file system are loaded by Pipenv as dotenv files.

    To prevent errors for happening we have to disable this feature.

    Additionally, it seems that env vars have to be changed before using
    venv, so an autouse fixture is required (cannot put this part in the
    beginning of the test function.
    """
    monkeypatch.setenv("PIPENV_DONT_LOAD_ENV", "1")
    monkeypatch.setenv("PIPENV_IGNORE_VIRTUALENVS", "1")
    monkeypatch.setenv("PIP_IGNORE_INSTALLED", "1")
    monkeypatch.setenv("PIPENV_VERBOSITY", "-1")


@pytest.mark.skipif(
    os.name == "nt", reason="pipenv fails due to colors (non-utf8) under Windows 10"
)
def test_pipenv_works_with_snek(tmpfolder, monkeypatch, venv):
    # Given a project is created with snek
    # and it has some dependencies in setup.cfg
    create_project(project_path="myproj", requirements=["platformdirs"])

    if any(ch in snek_version for ch in ("b", "a", "pre", "rc")):
        flags = "--pre"
    else:
        flags = ""

    with tmpfolder.join("myproj").as_cwd():
        # When we install pipenv,
        venv.run("pip install -v pipenv")

        try:
            venv.run("pipenv --bare install certifi")
            # use it to proxy setup.cfg
            venv.run(f"pipenv --bare install {flags} -e .")
            # and install things to the dev env,
            venv.run("pipenv --bare install --dev flake8")
            # Then it should be able to generate a Pipfile.lock
            venv.run("pipenv lock")
        except Exception:
            if sys.version_info[:2] <= (3, 6):
                # TODO: Remove try...except when 3.6 is no longer supported
                pytest.skip("Skip Pipenv specific problem for 3.6")
            else:
                raise

        assert Path("Pipfile.lock").exists()

        # with the correct dependencies
        with open("Pipfile.lock") as fp:
            content = json.load(fp)
            assert content["default"]["platformdirs"]
            assert content["develop"]["flake8"]

        # and run things from inside pipenv's venv
        pipenv_path = venv.run("pipenv --venv")
        assert pipenv_path in venv.run("pipenv run which flake8")
        venv.run("pipenv --bare run flake8 src/myproj/cli.py")


@pytest.mark.xfail(
    sys.version_info < (3, 7), reason="pip-compile may fail in old Python"
)
def test_piptools_works_with_snek(tmpfolder, monkeypatch):
    venv_path = Path(str(tmpfolder), "myproj/.venv").resolve()
    find = partial(find_venv_bin, venv_path)
    # Given a project is created with snek
    # and it has some dependencies in setup.cfg
    create_project(
        project_path="myproj", extensions=[Venv()], requirements=["platformdirs"]
    )
    with tmpfolder.join("myproj").as_cwd():
        requirements_in = Path("requirements.in")
        # When we install pip-tools
        run(f"{find('python')} -m pip install -v pip-tools certifi")
        # and write a requirements.in file that proxies setup.cfg
        # and install other things,
        requirements_in.write_text("-e file:.\nflake8")
        # Then we should be able to generate a requirements.txt
        run(find("pip-compile"))
        requirements_txt = Path("requirements.txt")
        assert requirements_txt.exists()
        # with the correct dependencies
        content = requirements_txt.read_text()
        assert "platformdirs==" in content
        assert "flake8==" in content
        assert "file:." in content
        # install the dependencies
        # and run things from inside pipenv's venv
        pip_sync = find("pip-sync")
        try:
            # pip-tools have problems on windows inside a test env with relative paths
            run(pip_sync)
            run(f"{find('flake8')} src/myproj/cli.py")
        except CalledProcessError as ex:
            if "assert" in ex.output:
                pytest.skip(
                    "pip-tools tries to assert a path is absolute, which fails "
                    "inside test env for some OSs"
                )
            else:
                raise



"""Internal library for manipulating, creating and dealing with names, or more generally
identifiers.
"""

import keyword
import re
from typing import Callable, Iterable, List, TypeVar

from .exceptions import InvalidIdentifier

T = TypeVar("T")


def is_valid_identifier(string: str) -> bool:
    """Check if string is a valid package name

    Args:
        string: package name

    Returns:
        True if string is valid package name else False
    """
    if not re.match("[_A-Za-z][_a-zA-Z0-9]*$", string):
        return False
    if keyword.iskeyword(string):
        return False
    return True


def make_valid_identifier(string: str) -> str:
    """Try to make a valid package name identifier from a string

    Args:
        string: invalid package name

    Returns:
        Valid package name as string or :obj:`RuntimeError`

    Raises:
        :obj:`InvalidIdentifier`: raised if identifier can not be converted
    """
    string = str(string).strip()
    string = string.replace("-", "_")
    string = string.replace(" ", "_")
    string = re.sub("[^_a-zA-Z0-9]", "", string)
    string = string.lower()
    if is_valid_identifier(string):
        return string

    raise InvalidIdentifier("String cannot be converted to a valid identifier.")


# from https://en.wikibooks.org/, Creative Commons Attribution-ShareAlike 3.0
def levenshtein(s1: str, s2: str) -> int:
    """Calculate the Levenshtein distance between two strings

    Args:
        s1: first string
        s2: second string

    Returns:
        Distance between s1 and s2
    """
    if len(s1) < len(s2):
        return levenshtein(s2, s1)

    # len(s1) >= len(s2)
    if len(s2) == 0:
        return len(s1)

    previous_row = list(range(len(s2) + 1))
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def dasherize(word: str) -> str:
    """Replace underscores with dashes in the string.

    Example::

        >>> dasherize("foo_bar")
        "foo-bar"

    Args:
        word (str): input word

    Returns:
        input word with underscores replaced by dashes
    """
    return word.replace("_", "-")


CAMEL_CASE_SPLITTER = re.compile(r"\W+|([A-Z][^A-Z\W]*)")


def underscore(word: str) -> str:
    """Convert CamelCasedStrings or dasherized-strings into underscore_strings.

    Example::

        >>> underscore("FooBar-foo")
        "foo_bar_foo"
    """
    return "_".join(w for w in CAMEL_CASE_SPLITTER.split(word) if w).lower()


def deterministic_name(obj):
    """Private API that returns an string that can be used to deterministically
    deduplicate and sort sequences of objects.
    """
    mod_name = getattr(obj, "__module__", "..")
    qual_name = getattr(obj, "__qualname__", obj.__class__.__qualname__)
    return f"{mod_name}.{qual_name}"


def deterministic_sort(sequence: Iterable[T]) -> List[T]:
    """Private API that order a sequence of objects lexicographically (by
    :obj:`deterministic_name`), removing duplicates, which is needed for determinism.

    The main purpose of this function is to deterministically sort a sequence of
    Snek extensions (it will also sort internal extensions before external:
    "snek.*" < "snekext.*").
    """
    deduplicated = {deterministic_name(x): x for x in sequence}
    # ^  duplicated keys will overwrite each other, so just one of them is left
    return [v for (_k, v) in sorted(deduplicated.items())]


def get_id(function: Callable) -> str:
    """Given a function, calculate its identifier.

    A identifier is a string in the format ``<module name>:<function name>``,
    similarly to the convention used for setuptools entry points.

    Note:
        This function does not return a Python 3 ``__qualname__`` equivalent.
        If the function is nested inside another function or class, the parent
        name is ignored.

    Args:
        function (callable): function object
    """
    return f"{function.__module__}:{function.__name__}"



"""
Extension that omits the creation of file `pyproject.toml`.

Since the isolated builds with PEP517/PEP518 are not completely backward compatible,
this extension provides an escape hatch for people that want to maintain the legacy
behaviour.

.. warning::
   This extension is **transitional** and will be removed in future versions of
   snek.
   Once support for isolated builds stabilises, the Python
   community will likely move towards using them more exclusively.
"""

from typing import List

from .. import structure
from ..actions import Action, ActionParams, ScaffoldOpts, Structure
from . import Extension


class NoPyProject(Extension):
    """Do not include a pyproject.toml file in the project root, and thus avoid isolated
    builds as defined in PEP517/518 [not recommended]
    """

    name = "no_pyproject"

    def activate(self, actions: List[Action]) -> List[Action]:
        actions = self.register(actions, ensure_option, before="get_default_options")
        return self.register(actions, remove_files, after="define_structure")


def ensure_option(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Make option available in non-CLI calls (used by other parts of snek)"""
    return struct, {**opts, "pyproject": False, "isolated_build": False}


def remove_files(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    struct = structure.reject(struct, "pyproject.toml")
    return struct, opts
